#!/usr/bin/env python
#
# volumeopts.py - Defines the VolumeOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module defines the :class:`VolumeOpts` class."""


import copy
import logging

import numpy as np

import fsl.data.image       as fslimage
import fsleyes_props        as props
import fsleyes.gl           as fslgl

import fsleyes.colourmaps   as fslcm
from . import colourmapopts as cmapopts
from . import volume3dopts  as vol3dopts
from . import                  niftiopts


log = logging.getLogger(__name__)


class VolumeOpts(cmapopts.ColourMapOpts,
                 vol3dopts.Volume3DOpts,
                 niftiopts.NiftiOpts):
    """The ``VolumeOpts`` class defines options for displaying :class:`.Image`
    instances as regular 3D volumes.
    """


    channel = props.Choice(('R', 'G', 'B', 'A'))
    """For images with the NIfTI ``RGB24`` or ``RGBA32`` data type,
    this property controls the channel that gets displayed.
    """


    clipImage = props.Choice()
    """Clip voxels according to the values in another image. By default, voxels
    are clipped by the values in the image itself - this property allows the
    user to choose another image by which voxels are to be clipped. Any image
    which is in the :class:`.OverlayList` can be selected for clipping. The
    :attr:`.ColourMapOpts.clippingRange` property dictates the values outside
    of which voxels are clipped.
    """


    modulateImage = props.Choice()
    """Modulate alapha (opacity) by the intensity of values in the selected
    image, instead of in this image. Only relevant when
    :attr:`.ColourMapOpts.modulateAlpha` is active.
    """


    interpolation = props.Choice(('none', 'linear', 'spline'))
    """How the value shown at a real world location is derived from the
    corresponding data value(s). ``none`` is equivalent to nearest neighbour
    interpolation.
    """


    @classmethod
    def getInitialDisplayRange(cls):
        """This class method returns a tuple containing ``(low, high)``
        percentile values which are used to set the initial values for the
        :attr:`.ColourMapOpts.displayRange` and
        :attr:`.ColourMapOpts.clippingRange` properties. If the initial
        display range has not yet been set (via the
        :meth:`setInitialDisplayRange` method), ``None`` is returned.
        """
        try:
            return cls.__initialDisplayRange
        except AttributeError:
            return None


    @classmethod
    def setInitialDisplayRange(cls, drange):
        """Sets the initial values for the :attr:`.ColourMapOpts.displayRange`
        and :attr:`.ColourMapOpts.clippingRange` to be used for new
        :class:`VolumeOpts` instances.

        :arg drange: A tuple containing ``(low, high)`` display range values
                     as percentiles of the image data range. May be ``None``,
                     in which case the initial display range will be set to the
                     image data range.
        """

        if drange is not None:
            low, high = drange
            if not all((low < high,
                        low >= 0,
                        low <= 100,
                        high >= 0,
                        high <= 100)):
                raise ValueError('Invalid initial display '
                                 'range: {}'.format(drange))

        cls.__initialDisplayRange = drange


    def __init__(self,
                 overlay,
                 display,
                 overlayList,
                 displayCtx,
                 **kwargs):
        """Create a :class:`VolumeOpts` instance for the specified ``overlay``,
        assumed to be an :class:`.Image` instance.

        All arguments are passed through to the :class:`.DisplayOpts`
        constructor.
        """

        # We need GL >= 2.1 for
        # spline interpolation
        if float(fslgl.GL_COMPATIBILITY) < 2.1:
            interp = self.getProp('interpolation')
            interp.removeChoice('spline', instance=self)
            interp.updateChoice('linear', instance=self, newAlt=['spline'])

        # Interpolation cannot be unbound
        # between VolumeOpts instances. This is
        # primarily to reduce memory requirement
        # - if interpolation were different
        # across different views, we would have
        # to create multiple 3D image textures
        # for the same image. Same goes for
        # clip/mod images
        nounbind = kwargs.get('nounbind', [])
        nounbind.append('interpolation')
        nounbind.append('clipImage')
        nounbind.append('modulateImage')
        kwargs['nounbind'] = nounbind

        # Some FSL tools will set the nifti aux_file
        # field to the name of a colour map - Check
        # to see if this is the case (do this before
        # calling __init__, so we don't clobber any
        # existing values).
        cmap = str(overlay.header.get('aux_file', 'none')).lower()

        if cmap == 'mgh-subcortical': cmap = 'subcortical'
        if cmap == 'mgh-cortical':    cmap = 'cortical'

        if cmap in fslcm.getColourMaps():
            self.cmap = cmap

        niftiopts.NiftiOpts.__init__(self,
                                     overlay,
                                     display,
                                     overlayList,
                                     displayCtx,
                                     **kwargs)

        # Some things only happen
        # on the parent instance
        self.__registered = self.getParent() is None

        # Check whether the data range for this
        # image is silly. If it is, and we are
        # on a platform that cannot use floating
        # point textures, we turn on the override
        # data range option.
        if self.__registered and np.issubdtype(overlay.dtype, np.floating):
            import fsleyes.gl.textures.data as texdata
            if not texdata.canUseFloatTextures()[0]:

                dmin, dmax = overlay.dataRange

                # a range of greater than 10e7
                # is defined as being "silly"
                if abs(dmax - dmin) > 10e7:

                    if   overlay.ndim == 3: sample = overlay[:]
                    elif overlay.ndim == 4: sample = overlay[..., 0]

                    drange = np.percentile(sample[sample != 0], [1, 99])

                    self.overrideDataRange       = drange
                    self.enableOverrideDataRange = True

        # Configure the initial display
        # range for new images, from the
        # initialDisplayRange percentiles.
        # We do this before ColourMapOpts.init
        drange = VolumeOpts.getInitialDisplayRange()

        if self.__registered and drange is not None:

            if   overlay.ndim == 3: sample = overlay[:]
            elif overlay.ndim == 4: sample = overlay[..., 0]

            drange = np.percentile(sample[sample != 0], drange)
            crange = [drange[0], overlay.dataRange[1]]

            self.displayRange  = drange
            self.modulateRange = drange
            self.clippingRange = crange

        # If this is not a RGB(A) image, disable
        # the channel property. If it's a RGB
        # image,  remove the "A" option from
        # the channel property.
        if self.__registered:

            nchannels = self.overlay.nvals
            if nchannels == 1:
                self.disableProperty('channel')
            elif nchannels == 3:
                prop = self.getProp('channel')
                prop.removeChoice('A', self)

        cmapopts .ColourMapOpts.__init__(self)
        vol3dopts.Volume3DOpts .__init__(self)

        # Both parent and child VolumeOpts instances
        # listen for Image dataRange changes. The data
        # range for large images may be calculated
        # asynchronously on a separate thread, meaning
        # that data range updates may occur at random
        # times.
        #
        # If parent instances did not listen for data
        # range updates and, at startup, the following
        # sequence of events occurs:
        #
        #   1. Parent VolumeOpts instance created
        #
        #   2. Image.dataRange updated
        #
        #   3. Child VolumeOpts instance created
        #
        # The known parent data range will be 0-0,
        # the child will not receive any notification
        # about the data range change, and the child's
        # data range will be clobbered by the parent's.
        # This ugly situation is avoided simply by
        # having the parent track changes to the data
        # range in addition to all children.
        overlay.register(self.name,
                         self.__dataRangeChanged,
                         'dataRange',
                         runOnIdle=True)

        # We need to listen for changes to clipImage
        # and to [enable]overrideDataRange, as they
        # will change the display data range. These
        # cannot be unbound between parent/children,
        # so only the parent needs to listen.
        if self.__registered:
            overlayList.addListener('overlays',
                                    self.name,
                                    self.__overlayListChanged)
            self       .addListener('clipImage',
                                    self.name,
                                    self.__clipImageChanged)
            self       .addListener('modulateImage',
                                    self.name,
                                    self.__modulateImageChanged)
            self       .addListener('enableOverrideDataRange',
                                    self.name,
                                    self.__enableOverrideDataRangeChanged)
            self       .addListener('overrideDataRange',
                                    self.name,
                                    self.__overrideDataRangeChanged)

            self.__overlayListChanged()
            self.__clipImageChanged(    updateDataRange=False)
            self.__modulateImageChanged(updateDataRange=False)


    def destroy(self):
        """Removes property listeners, and calls the :meth:`NiftiOpts.destroy`
        method.
        """

        overlay     = self.overlay
        overlayList = self.overlayList

        overlay.deregister(self.name, 'dataRange')

        if self.__registered:

            overlayList.removeListener('overlays',                self.name)
            self       .removeListener('clipImage',               self.name)
            self       .removeListener('modulateImage',           self.name)
            self       .removeListener('enableOverrideDataRange', self.name)
            self       .removeListener('overrideDataRange',       self.name)

        cmapopts .ColourMapOpts.destroy(self)
        vol3dopts.Volume3DOpts .destroy(self)
        niftiopts.NiftiOpts    .destroy(self)


    def getDataRange(self):
        """Overrides :meth:`.ColourMapOpts.getDataRange`. Returns the
        :attr:`.Image.dataRange` of the image, or the
        :attr:`overrideDataRange` if it is active.
        """
        if self.enableOverrideDataRange: return self.overrideDataRange
        else:                            return self.overlay.dataRange


    def getClippingRange(self):
        """Overrides :meth:`.ColourMapOpts.getClippingRange`.
        If a :attr:`.clipImage` is set, returns its data range. Otherwise
        returns ``None``.
        """

        if self.clipImage is None:
            return cmapopts.ColourMapOpts.getClippingRange(self)
        else:
            return self.clipImage.dataRange


    def getModulateRange(self):
        """Overrides :meth:`.ColourMapOpts.getModulateRange`.
        If a :attr:`.modulateImage` is set, returns its data range. Otherwise
        returns ``None``.
        """

        if self.modulateImage is None:
            return cmapopts.ColourMapOpts.getModulateRange(self)
        else:
            return self.modulateImage.dataRange


    def __dataRangeChanged(self, *a):
        """Called when the :attr:`.Image.dataRange` property changes.
        Calls :meth:`.ColourMapOpts.updateDataRange`.
        """
        self.updateDataRange(False, False, False)


    def __enableOverrideDataRangeChanged(self, *a):
        """Called when the :attr:`enableOverrideDataRange` property changes.
        Calls :meth:`.ColourMapOpts.updateDataRange`.
        """
        self.updateDataRange()


    def __overrideDataRangeChanged(self, *a):
        """Called when the :attr:`overrideDataRange` property changes.
        Calls :meth:`.ColourMapOpts.updateDataRange`.
        """
        self.updateDataRange()


    def __overlayListChanged(self, *a):
        """Called when the :`class:`.OverlayList` changes. Updates the
        options of the :attr:`clipImage` property.
        """

        clipProp = self.getProp('clipImage')
        clipVal  = self.clipImage
        modProp  = self.getProp('modulateImage')
        modVal   = self.modulateImage
        overlays = self.displayCtx.getOrderedOverlays()

        options  = [None]

        for overlay in overlays:

            if overlay is self.overlay:                 continue
            if not isinstance(overlay, fslimage.Image): continue

            options.append(overlay)

        clipProp.setChoices(options, instance=self)
        modProp .setChoices(options, instance=self)

        if clipVal in options: self.clipImage     = clipVal
        else:                  self.clipImage     = None
        if modVal  in options: self.modulateImage = modVal
        else:                  self.modulateImage = None


    def __clipImageChanged(self, *a, **kwa):
        """Called when the :attr:`clipImage` property is changed. Updates
         the range of the :attr:`clippingRange` property.

        :arg updateDataRange: Defaults to ``True``. If ``False``, the
                              :meth:`.ColourMapOpts.updateDataRange` method
                              is not called.
        """

        updateDR = kwa.get('updateDataRange', True)

        haveClipImage = self.clipImage is not None

        if not haveClipImage:
            self.enableProperty('linkLowRanges')
            self.enableProperty('linkHighRanges')

        # If the clipping range is based on another
        # image, it makes no sense to link the low/
        # high display/clipping ranges, as they are
        # probably different. So if a clip image is
        # selected, we disable the link range
        # properties.
        elif self.propertyIsEnabled('linkLowRanges'):

            self.linkLowRanges  = False
            self.linkHighRanges = False

            self.disableProperty('linkLowRanges')
            self.disableProperty('linkHighRanges')

        log.debug('Clip image changed for %s: %s',
                  self.overlay, self.clipImage)

        if updateDR:
            self.updateDataRange(resetDR=False, resetMR=False)


    def __modulateImageChanged(self, *a, **kwa):
        """Called when the :attr:`modulateImage` property is changed. Updates
         the range of the :attr:`modulateRange` property.

        :arg updateDataRange: Defaults to ``True``. If ``False``, the
                              :meth:`.ColourMapOpts.updateDataRange` method
                              is not called.
        """

        updateDR = kwa.get('updateDataRange', True)

        log.debug('Modulate image changed for %s: %s',
                  self.overlay, self.modulateImage)

        if updateDR:
            self.updateDataRange(resetDR=False, resetCR=False)


class VolumeRGBOpts(niftiopts.NiftiOpts):
    """The ``VolumeRGBOpts`` class is intended for displaying
    :class:`.Image` instances containing RGB(A) data.
    """


    rColour = props.Colour(default=(1, 0, 0))
    """Colour to use for the red channel. """


    gColour = props.Colour(default=(0, 1, 0))
    """Colour to use for the green channel. """


    bColour = props.Colour(default=(0, 0, 1))
    """Colour to use for the blue channel. """


    suppressR = props.Boolean(default=False)
    """Suppress the R channel. """


    suppressG = props.Boolean(default=False)
    """Suppress the G channel. """


    suppressB = props.Boolean(default=False)
    """Suppress the B channel. """


    suppressA = props.Boolean(default=False)
    """Suppress the A channel. """


    suppressMode = props.Choice(('white', 'black', 'transparent'))
    """How colours should be suppressed. """


    interpolation = copy.copy(VolumeOpts.interpolation)
    """See :attr:`VolumeOpts.interpolation`. """


    def __init__(self,
                 overlay,
                 display,
                 overlayList,
                 displayCtx,
                 **kwargs):
        """Create a :class:`VolumeRGBOpts` instance for the specified
        ``overlay``, assumed to be an :class:`.Image` instance with type
        ``NIFTI_TYPE_RGB24`` or ``NIFTI_TYPE_RGBA32``.

        All arguments are passed through to the :class:`.DisplayOpts`
        constructor.
        """

        # We need GL >= 2.1 for
        # spline interpolation
        if float(fslgl.GL_COMPATIBILITY) < 2.1:
            interp = self.getProp('interpolation')
            interp.removeChoice('spline', instance=self)
            interp.updateChoice('linear', instance=self, newAlt=['spline'])

        niftiopts.NiftiOpts.__init__(self,
                                     overlay,
                                     display,
                                     overlayList,
                                     displayCtx,
                                     **kwargs)



class ComplexOpts(VolumeOpts):
    """The ``ComplexOpts`` class is a specialisation of :class:`VolumeOpts` for
    images with a complex data type.
    """


    component = props.Choice(('real', 'imag', 'mag', 'phase'))
    """How to display the complex data:

     - ``'real'``   - display the real component
     - ``'imag'```  - display the imaginary component
     - ``'mag'```   - display the magnitude
     - ``'phase'``` - display the phase
    """


    def __init__(self, *args, **kwargs):
        """Create a ``ComplexOpts``. All arguments are passed through to
        the :class:`VolumeOpts` constructor.
        """
        self.__dataRanges = {}
        VolumeOpts.__init__(self, *args, **kwargs)
        self.addListener('component', self.name, self.__componentChanged)


    def destroy(self):
        """Must be called when this ``ComplexOpts`` is no longer needed. """
        VolumeOpts.destroy(self)


    def getDataRange(self):
        """Overrides :meth:`.ColourMapOpts.getDataRange`.
        Calculates and returns the data range of the current
        :attr:`component`.
        """

        drange = self.__dataRanges.get(self.component, None)
        if drange is None:
            data   = self.getComponent(self.overlay[:])
            drange = np.nanmin(data), np.nanmax(data)
            self.__dataRanges[self.component] = drange
        return drange


    def getComponent(self, data):
        """Calculates and returns the current :attr:`component` from the given
        data, assumed to be complex.
        """
        if   self.component == 'real':  return self.getReal(data)
        elif self.component == 'imag':  return self.getImaginary(data)
        elif self.component == 'mag':   return self.getMagnitude(data)
        elif self.component == 'phase': return self.getPhase(data)


    @staticmethod
    def getReal(data):
        """Return the real component of the given complex data. """
        return data.real


    @staticmethod
    def getImaginary(data):
        """Return the imaginary component of the given complex data. """
        return data.imag


    @staticmethod
    def getMagnitude(data):
        """Return the magnitude of the given complex data. """
        return (data.real ** 2 + data.imag ** 2) ** 0.5


    @staticmethod
    def getPhase(data):
        """Return the phase of the given complex data. """
        return np.arctan2(data.imag, data.real)


    def __componentChanged(self, *a):
        """Called when the :attr:`component` changes. Calls
        :meth:`.ColourMapOpts.updateDataRange`.
        """
        self.updateDataRange()



#!/usr/bin/env python
#
# wxglscene3dcanvas.py - The WXGLScene3DCanvas class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`WXGLScene3DCanvas` class, which is a
:class:`.Scene3DCanvas` for use in a :mod:`wx` application.
"""


import six

import wx.glcanvas as wxgl

import fsleyes.gl               as fslgl
import fsleyes.gl.scene3dcanvas as scene3dcanvas


class WXGLScene3DCanvas(six.with_metaclass(fslgl.WXGLMetaClass,
                                           scene3dcanvas.Scene3DCanvas,
                                           fslgl.WXGLCanvasTarget,
                                           wxgl.GLCanvas)):
    """The ``WXGLScene3DCanvas`` is a :class:`.Scene3DCanvas`, a
    :class:`wx.glcanvas.GLCanvas` and a :class:`.WXGLCanvasTarget`. If you
    want to use a :class:`.Scene3DCanvas` in your :mod:`wx` application, then
    you should use a ``WXGLScene3DCanvas``.
    """


    def __init__(self, parent, overlayList, displayCtx):
        """Create a ``WXGLScene3DCanvas``. See :meth:`.Scene3DCanvas.__init__`
        for details on the arguments.
        """
        attrs = fslgl.WXGLCanvasTarget.displayAttribues()

        wxgl.GLCanvas              .__init__(self, parent, **attrs)
        fslgl.WXGLCanvasTarget     .__init__(self)
        scene3dcanvas.Scene3DCanvas.__init__(self, overlayList, displayCtx)


    def destroy(self):
        """Must be called when this ``WXGLScene3DCanvas`` is no longer needed.
        Clears some event listeners and calls the base class ``destroy``
        method.
        """
        scene3dcanvas.Scene3DCanvas.destroy(self)
        fslgl.WXGLCanvasTarget     .destroy(self)



#!/usr/bin/env python
#
# colourbar.py - The ColourBarPanel.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ColourBarPanel`, a :class:`.FSLeyesPanel`
which renders a colour bar.
"""


import wx

import fsleyes.panel                  as fslpanel
import fsleyes.gl.wxglcolourbarcanvas as cbarcanvas
import fsleyes.controls.colourbar     as cbar


class ColourBarPanel(fslpanel.FSLeyesPanel):
    """The ``ColourBarPanel`` is a panel which shows a colour bar, depicting
    the data range of the currently selected overlay (if applicable). A
    :class:`.ColourBarCanvas` is used to render the colour bar.


    .. note:: Currently, the ``ColourBarPanel`` will only display a colour bar
              for overlays which are associated with a :class:`.ColourMapOpts`
              instance.
    """


    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create a ``ColourBarPanel``.

        :arg parent:      The :mod:`wx` parent object.

        :arg overlayList: The :class:`.OverlayList` instance.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg frame:       The :class:`.FSLeyesFrame`.
        """

        fslpanel.FSLeyesPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        self.__cbCanvas = cbarcanvas.WXGLColourBarCanvas(
            self, overlayList, displayCtx)

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(self.__sizer)
        self.__sizer.Add(self.__cbCanvas, flag=wx.EXPAND, proportion=1)

        self.__cbCanvas.colourBar.addListener(
            'orientation', self.name, self.__layout)
        self.__cbCanvas.colourBar.addListener(
            'fontSize', self.name, self.__layout)
        self.__layout()


    def getCanvas(self):
        """Returns the :class:`.ColourBarCanvas` which displays the rendered
        colour bar.
        """
        return self.__cbCanvas


    @property
    def canvas(self):
        """Returns the :class:`.ColourBarCanvas` which displays the rendered
        colour bar.
        """
        return self.__cbCanvas


    @property
    def colourBar(self):
        """Returns the :class:`.ColourBar` which generates the colour bar
        bitmap.
        """
        return self.__cbCanvas.colourBar


    def destroy(self):
        """Must be called when this ``ColourBarPanel`` is no longer needed. """

        self.__cbCanvas.colourBar.removeListener('orientation', self.name)
        self.__cbCanvas.colourBar.removeListener('fontSize',    self.name)
        self.__cbCanvas.destroy()
        self.__cbCanvas = None

        fslpanel.FSLeyesPanel.destroy(self)


    def __layout(self, *a):
        """Called when this ``ColourBarPanel`` needs to be laid out.
        Sets the panel size, and calls the :meth:`__refreshColourBar` method.
        """
        canvas   = self.__cbCanvas
        w, h     = self.GetClientSize().Get()
        cbarSize = cbar.colourBarMinorAxisSize(canvas.colourBar.fontSize)

        if canvas.colourBar.orientation == 'horizontal':
            canvas.SetSizeHints(-1, cbarSize, -1, cbarSize, -1, -1)
        else:
            canvas.SetSizeHints(cbarSize, -1, cbarSize, -1, -1, -1)

        wx.CallAfter(self.Layout)



#!/usr/bin/env python
#
# moviegif.py - The MovieGifAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`MovieGifAction`, which allows the user
to save animated gifs. The :func:`makeGif` function can also be used to
programmatically generate animated gifs.
"""


import os.path as op
import            os
import            shutil
import            tempfile

import            PIL
import            wx
import numpy   as np

import fsl.utils.idle                 as idle
import fsl.transform.affine           as affine
import fsl.utils.settings             as fslsettings
import fsleyes_widgets.utils.progress as progress
from   fsleyes_widgets import            isalive

from . import base

import fsleyes.strings            as strings
import fsleyes.actions.screenshot as screenshot
import fsleyes.views.scene3dpanel as scene3dpanel


class MovieGifAction(base.Action):
    """The ``MovieGifAction`` allows the user to save an animated gif of the
    currently selected overlay in a :class:`.CanvasPanel`, according to the
    current movie mode settings.
    """

    def __init__(self, overlayList, displayCtx, panel):
        """Create a ``MovieGifAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg panel:       The :class:`.CanvasPanel` to generate the animated
                          GIF for.
        """

        base.Action.__init__(self, overlayList, displayCtx, self.__doMakeGif)

        self.__name  = '{}_{}'.format(type(self).__name__, id(self))
        self.__panel = panel

        overlayList.addListener('overlays',
                                self.__name,
                                self.__selectedOverlayChanged)
        displayCtx .addListener('selectedOverlay',
                                self.__name,
                                self.__selectedOverlayChanged)
        panel      .addListener('movieAxis',
                                self.__name,
                                self.__selectedOverlayChanged)

        self.__selectedOverlayChanged()


    def destroy(self):
        """Must be called when this ``MovieGifAction`` is no longer neded.
        Removes some property listeners.
        """
        self.overlayList.removeListener('overlays',        self.__name)
        self.displayCtx .removeListener('selectedOverlay', self.__name)
        self.__panel    .removeListener('movieAxis',       self.__name)
        base.Action.destroy(self)


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` changes.
        Enables/disables this action based on whether a movie can be played
        (see :meth:`.CanvasPanel.canRunMovie`).
        """

        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is None:
            self.enabled = False
            return

        opts = self.displayCtx.getOpts(overlay)

        self.enabled = self.__panel.canRunMovie(overlay, opts)


    def __doMakeGif(self):
        """Prompts the user to select a file to save the movie to, and then
        generates the movie via :func:`makeGif`.
        """

        lastDirSetting = 'fsleyes.actions.screenshot.lastDir'
        filename       = 'movie.gif'
        fromDir        = fslsettings.read(lastDirSetting, os.getcwd())

        dlg = wx.FileDialog(self.__panel,
                            message=strings.messages[self, 'movieGif'],
                            defaultDir=fromDir,
                            defaultFile=filename,
                            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)

        if dlg.ShowModal() != wx.ID_OK:
            return

        filename = dlg.GetPath()
        self.__progdlg  = progress.Bounce(
            'Generating GIF',
            'Generating GIF',
            style=wx.PD_APP_MODAL | wx.PD_CAN_ABORT,
            parent=self.__panel)

        def update(frame):
            if self.__progdlg is not None and isalive(self.__progdlg):
                self.__progdlg.DoBounce('Saved frame {}...'.format(frame))
                return not self.__progdlg.WasCancelled()
            else:
                return False

        def finish():
            if isalive(self.__progdlg):
                self.__progdlg.Hide()
                self.__progdlg.Close()
            self.__progdlg = None

        # TODO prompt user to select axis/delay/limits?
        self.__progdlg.Show()
        makeGif(self.overlayList,
                self.displayCtx,
                self.__panel,
                filename,
                progfunc=update,
                onfinish=finish)


def makeGif(overlayList,
            displayCtx,
            panel,
            filename,
            progfunc=None,
            onfinish=None):
    """Save an animated gif of the currently selected overlay, according to the
    current movie mode settings.

    .. note:: This function will return immediately, as the animated GIF is
              generated on the ``wx`` :mod:`.idle` loop

    :arg overlayList: The :class:`.OverlayList`
    :arg displayCtx:  The :class:`.DisplayContext`
    :arg panel:       The :class:`.CanvasPanel`.
    :arg filename:    Name of file to save the movie to
    :arg progfunc:    Function which will be called after each frame is saved.
    :arg onfinish:    Function which will be called after all frames have been
                      saved.
    """

    def defaultProgFunc(frame):
        return True

    if progfunc is None:
        progfunc = defaultProgFunc

    overlay = displayCtx.getSelectedOverlay()
    opts    = displayCtx.getOpts(overlay)
    tempdir = tempfile.mkdtemp()
    is3d    = isinstance(panel, scene3dpanel.Scene3DPanel) and \
              panel.movieAxis != 3

    class Context(object):
        pass

    ctx           = Context()
    ctx.cancelled = False
    ctx.images    = []
    ctx.frames    = []

    class Finished(Exception):
        pass

    class Cancelled(Exception):
        pass

    def finalise(ctx):
        if not ctx.cancelled and len(ctx.images) > 0:
            ctx.images[0].save(filename,
                               format='gif',
                               save_all=True,
                               append_images=ctx.images[1:],
                               duration=50,
                               loop=0)
        shutil.rmtree(tempdir)

        if onfinish is not None:
            onfinish()

    def ready():
        globjs = [c.getGLObject(o)
                  for c in panel.getGLCanvases()
                  for o in overlayList]
        globjs = [g for g in globjs if g is not None]
        return all([g.ready() for g in globjs])

    def captureFrame(ctx):
        try:
            panel.movieSync()
            realCaptureFrame(ctx)
            idle.idleWhen(captureFrame, ready, ctx, after=0.1)
            panel.doMovieUpdate(overlay, opts)

        except Finished:
            finalise(ctx)

        except (Cancelled, Exception) as e:
            ctx.cancelled = True
            finalise(ctx)

    def realCaptureFrame(ctx):

        # individual frames don't
        # need to be saved as gif
        idx   = len(ctx.frames)
        fname = op.join(tempdir, '{}.png'.format(idx))

        frame = panel.getMovieFrame(overlay, opts)

        if not progfunc(idx):
            raise Cancelled()

        # The 3D X/Y/Z movie mode performs
        # rotations, rather than moving the
        # display location through the X/Y/Z
        # axes. The "frame" returned by
        # getMovieFrame is a rotation matrix.
        # We convert these rotation matrices
        # into rms-deviations (average
        # deviation of the current frame from
        # the starting frame), which has an
        # inverted "V"-shaped wave form as the
        # scene is rotated 360 degrees. So
        # we continue capturing frames until
        # the rmsdev of the current frame is:
        #
        #   - close to 0 (i.e. very similar to
        #     the rotation matrix of the starting
        #     frame), and
        #
        #   - less than the most recent frame (i.e.
        #     has rotated past 180 degrees, and is
        #     rotating back twoards the starting
        #     point)
        if is3d:

            if len(ctx.frames) == 0:
                ctx.startFrame = frame

            # normalise the rotmat for this
            # frame to the rms difference
            # from the starting rotmat
            frame = affine.rmsdev(ctx.startFrame, frame)

            # Keep capturing frames until we
            # have performed a full 360 degree
            # rotation (rmsdev of current
            # frame is decreasing towards 0)
            if len(ctx.frames) > 1    and \
               frame < ctx.frames[-1] and \
               abs(frame) < 0.1:
                raise Finished()

        # All other movie frames have a range
        # (fmin, fmax) and start at some arbitrary
        # point within this range. We capture frames
        # until a full loop through this range has
        # been completed.
        elif len(ctx.frames) > 1 and np.isclose(frame, ctx.frames[0]):
            raise Finished()

        screenshot.screenshot(panel, fname)
        ctx.images.append(PIL.Image.open(fname))
        ctx.frames.append(frame)

    idle.idleWhen(captureFrame, ready, ctx, after=0.1)



#!/usr/bin/env python
#
# offscreenslicecanvas.py - The OffScreenSliceCanvas.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OffScreenSliceCanvas`, which is a
:class:`.SliceCanvas` for use in off-screen OpenGL rendering.
"""


import fsleyes.gl         as fslgl
from . import slicecanvas as sc


class OffScreenSliceCanvas(sc.SliceCanvas, fslgl.OffScreenCanvasTarget):
    """The ``OffScreenSliceCanvas`` is a :class:`.SliceCanvas` which uses
    a :class:`.RenderTexture` as its target, for static off-screen OpenGL
    rendering.
    """

    def __init__(self,
                 overlayList,
                 displayCtx,
                 zax,
                 *args,
                 **kwargs):
        """Create an ``OffScreenSliceCanvas``.

        See the :class:`.SliceCanvas` and :class:`.OffscreenCanvasTarget`
        classes for details on the other arguments.
        """

        fslgl.OffScreenCanvasTarget.__init__(self, *args, **kwargs)
        sc.SliceCanvas             .__init__(self,
                                             overlayList,
                                             displayCtx,
                                             zax)



#!/usr/bin/env python
#
# program.py - The ARBPShader class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ARBPShader` class, which encapsulates
an OpenGL shader program written according to the ``ARB_vertex_program``
and ``ARB_fragment_program`` extensions.
"""


from __future__ import division

import re
import logging

import numpy                          as np

import OpenGL.GL                      as gl
import OpenGL.raw.GL._types           as gltypes
import OpenGL.GL.ARB.fragment_program as arbfp
import OpenGL.GL.ARB.vertex_program   as arbvp

import fsl.utils.memoize              as memoize
from . import                            parse


log = logging.getLogger(__name__)


class ARBPShader(object):
    """The ``ARBPShader`` class encapsulates an OpenGL shader program
    written according to the ``ARB_vertex_program`` and
    ``ARB_fragment_program`` extensions. It parses and compiles vertex
    and fragment program code, and provides methods to load/unload
    the program, and to set vertex/fragment program parameters and vertex
    attributes.


    The ``ARBPShader`` class assumes that vertex/fragment program source
    has been written to work with the functions defined in the
    :mod:`.arbp.parse` module, which allows programs to be written so that
    parameter, vertex attribute and texture locations do not have to be hard
    coded in the source.  Texture locations may be specified in
    :meth:`__init__`, and parameter/vertex attribute locations are
    automatically assigned by the ``ARBPShader``.


    The following methods are available on an ``ARBPShader`` instance:

    .. autosummary::
       :nosignatures:

       load
       unload
       destroy
       recompile
       setVertParam
       setFragParam
       setAtt
       setConstant

    Typcical usage of an ``ARBPShader`` will look something like the
    following::

        vertSrc = 'vertex shader source'
        fragSrc = 'vertex shader source'

        # You must specify the texture unit
        # assignments at creation time.
        textures = {
            'colourMapTexture' : 0,
            'dataTexture'      : 1
        }

        program = ARBPShader(vertSrc, fragSrc, textures)

        # Load the program, and
        # enable program attributes
        # (texture coordinates)
        program.load()
        progra.loadAtts()

        # Set some parameters
        program.setVertParam('transform', np.eye(4))
        program.setFragParam('clipping',  [0, 1, 0, 0])

        # Create and set vertex attributes
        vertices, normals = createVertices()

        program.setAtt('normals', normals)

        # Draw the scene
        gl.glDrawArrays(gl.GL_TRIANGLES, 0, len(vertices))

        # Clear the GL state
        program.unloadAtts()
        program.unload()

        # Delete the program when
        # it is no longer needed
        program.destroy()


    .. warning:: The ``ARBPShader`` uses texture coordinates to pass vertex
                 attributes to the shader programs. Therefore, if you are using
                 an ``ARBPShader`` you cannot directly use texture coordinates.


    See also the :class:`.GLSLShader`, which provides similar functionality for
    GLSL shader programs.
    """


    def __init__(self,
                 vertSrc,
                 fragSrc,
                 includePath,
                 textureMap=None,
                 constants=None,
                 clean=True):
        """Create an ``ARBPShader``.

        :arg vertSrc:     Vertex program source.

        :arg fragSrc:     Fragment program source.

        :arg textureMap:  A dictionary of ``{name : int}`` mappings, specifying
                          the texture unit assignments.

        :arg constants:   A dictionary of ``{name : values}`` mappings,
                          specifying any constant parameters required by the
                          programs. It is assumed that constant parameters are
                          shared by the vertex and fragment programs.

        :arg includePath: Path to a directory which contains any additional
                          files that may be included in the given source
                          files.

        :arg clean:       If ``True`` (the default), the vertex and fragment
                          program source is 'cleaned' before compilation - all
                          comments, empty lines, and unncessary spaces are
                          removed before compilation.
        """

        decs = parse.parseARBP(vertSrc, fragSrc)

        vParams      = decs['vertParam']
        fParams      = decs['fragParam']
        constantDecs = decs['constant']

        if constants is None: constants      = {}
        if len(vParams) > 0:  vParams, vLens = zip(*vParams)
        else:                 vParams, vLens = [], []
        if len(fParams) > 0:  fParams, fLens = zip(*fParams)
        else:                 fParams, fLens = [], []

        vLens = {name : length for name, length in zip(vParams, vLens)}
        fLens = {name : length for name, length in zip(fParams, fLens)}

        self.vertexSource    = vertSrc
        self.fragmentSource  = fragSrc
        self.includePath     = includePath
        self.vertexProgram   = None
        self.fragmentProgram = None
        self.clean           = clean
        self.includePath     = includePath
        self.vertParams      = vParams
        self.vertParamLens   = vLens
        self.fragParams      = fParams
        self.fragParamLens   = fLens
        self.textures        = decs['texture']
        self.attrs           = decs['attr']
        self.constants       = constantDecs
        self.constantVals    = dict(constants)

        # See the setAtt method for
        # information about this dict
        self.__attCache = {}

        poses = self.__generatePositions(textureMap)
        vpPoses, fpPoses, texPoses, attrPoses = poses

        self.vertParamPositions = vpPoses
        self.fragParamPositions = fpPoses
        self.texturePositions   = texPoses
        self.attrPositions      = attrPoses

        self.recompile()

        log.debug('{}.init({})'.format(type(self).__name__, id(self)))


    def __del__(self):
        """Prints a log message. """
        if log:
            log.debug('{}.del({})'.format(type(self).__name__, id(self)))


    def destroy(self):
        """Deletes all GL resources managed by this ``ARBPShader``. """

        if self.vertexProgram is not None:
            arbvp.glDeleteProgramsARB(1, gltypes.GLuint(self.vertexProgram))
        if self.fragmentProgram is not None:
            arbfp.glDeleteProgramsARB(1, gltypes.GLuint(self.fragmentProgram))

        self.vertexProgram   = None
        self.fragmentProgram = None


    def recompile(self):
        """(Re-)generates the vertex and fragment program source code, and
        recompiles the programs.
        """

        # As we are compiling new
        # vertex/fragment programs,
        # we need to invalidate any
        # cached parameter values.
        # Constants are ok.
        self.setVertParam.invalidate()
        self.setFragParam.invalidate()

        vertSrc, fragSrc = parse.fillARBP(self.vertexSource,
                                          self.fragmentSource,
                                          self.vertParamPositions,
                                          self.vertParamLens,
                                          self.fragParamPositions,
                                          self.fragParamLens,
                                          self.constantVals,
                                          self.texturePositions,
                                          self.attrPositions,
                                          self.includePath)

        # Compile the new version, but
        # only discard the old version
        # if compilation succeeds
        vp, fp = self.__compile(vertSrc, fragSrc)

        self.destroy()
        self.vertexProgram   = vp
        self.fragmentProgram = fp


    def load(self):
        """Loads the shader program. """
        gl.glEnable(arbvp.GL_VERTEX_PROGRAM_ARB)
        gl.glEnable(arbfp.GL_FRAGMENT_PROGRAM_ARB)

        arbvp.glBindProgramARB(arbvp.GL_VERTEX_PROGRAM_ARB,
                               self.vertexProgram)
        arbfp.glBindProgramARB(arbfp.GL_FRAGMENT_PROGRAM_ARB,
                               self.fragmentProgram)

    def loadAtts(self):
        """Enables texture coordinates for all shader program attributes. """
        for attr in self.attrs:
            texUnit = self.__getAttrTexUnit(attr)

            gl.glClientActiveTexture(texUnit)
            gl.glEnableClientState(gl.GL_TEXTURE_COORD_ARRAY)


    def unload(self):
        """Unloads the shader program. """
        gl.glDisable(arbfp.GL_FRAGMENT_PROGRAM_ARB)
        gl.glDisable(arbvp.GL_VERTEX_PROGRAM_ARB)


    def unloadAtts(self):
        """Disables texture coordinates on all texture units. """
        for attr in self.attrs:
            texUnit = self.__getAttrTexUnit(attr)

            gl.glClientActiveTexture(texUnit)
            gl.glDisableClientState(gl.GL_TEXTURE_COORD_ARRAY)

        self.__attCache = {}


    @memoize.Instanceify(memoize.skipUnchanged)
    def setVertParam(self, name, value):
        """Sets the value of the specified vertex program parameter.

        .. note:: It is assumed that the value is either a sequence of length
                  4 (for vector parameters), or a ``numpy`` array of shape
                  ``(n, 4)`` (for matrix parameters).

        .. note:: This method is decorated by the
                  :func:`.memoize.skipUnchanged` decorator, which returns
                  ``True`` if the value was changed, and ``False`` otherwise.
        """

        pos   = self.vertParamPositions[name]
        value = self.__normaliseParam(value)
        nrows = len(value) // 4

        log.debug('Setting vertex parameter {} = {}'.format(name, value))

        for i in range(nrows):
            row = value[i * 4: i * 4 + 4]
            arbvp.glProgramLocalParameter4fARB(
                arbvp.GL_VERTEX_PROGRAM_ARB, pos + i,
                row[0], row[1], row[2], row[3])


    @memoize.Instanceify(memoize.skipUnchanged)
    def setFragParam(self, name, value):
        """Sets the value of the specified vertex program parameter. See
        :meth:`setVertParam` for infomration about possible values.

        .. note:: This method is decorated by the
                  :func:`.memoize.skipUnchanged` decorator, which returns
                  ``True`` if the value was changed, and ``False`` otherwise.
        """
        pos   = self.fragParamPositions[name]
        value = self.__normaliseParam(value)
        nrows = len(value) // 4

        log.debug('Setting fragment parameter {} = {}'.format(name, value))

        for i in range(nrows):
            row = value[i * 4: i * 4 + 4]
            arbfp.glProgramLocalParameter4fARB(
                arbfp.GL_FRAGMENT_PROGRAM_ARB, pos + i,
                row[0], row[1], row[2], row[3])


    @memoize.Instanceify(memoize.skipUnchanged)
    def setConstant(self, name, value):
        """Updates the value of a constant parameter used by the program.

        The :meth:`recompile` method must be called after changing a constant
        value.
        """
        if name not in self.constants:
            raise ValueError('Unknown constant: {}'.format(name))

        log.debug('Setting vertex constant {} = {}'.format(name, value))

        self.constantVals[name] = value


    def setAtt(self, name, value):
        """Sets the value of the specified vertex attribute. Each vertex
        attribute is mapped to a texture coordinate. It is assumed that
        the given value is a ``numpy`` array of shape ``(n, l)``, where
        ``n`` is the number of vertices being drawn, and ``l`` is the
        number of components in each vertex attribute coordinate.
        """
        texUnit = self.__getAttrTexUnit(name)
        size    = value.shape[1]
        value   = np.array(value, dtype=np.float32, copy=False)

        log.debug('Setting vertex attribute {} [{}] = [{} * {}]'.format(
            name, texUnit, value.shape[0], size))

        # We must save a ref to the value so
        # that it doesn't get GC'd by python
        # before actually being used by GL.
        # This took me an entire day to
        # figure out. The cache gets cleared
        # on every call to unloadAtts.
        value = value.ravel('C')
        self.__attCache[name] = value

        gl.glClientActiveTexture(texUnit)
        gl.glTexCoordPointer(size, gl.GL_FLOAT, 0, value)


    def __normaliseParam(self, value):
        """Used by :meth:`setVertParam` and :meth:`setFragParam`. Ensures that
        all vertex/fragment program parameters are vectors of length 4, or
        matrices of size ``(n, 4)``.
        """

        # scalar
        if np.isscalar(value):
            value = [value]

        value = np.array(value, copy=False)

        # vector
        if len(value.shape) == 1:

            # if < 4 values, pad it to 4. If > 4
            # values, an error will be raised below
            if value.shape[0] < 4:
                value = list(value) + [0] * (4 - len(value))

        value = np.array(value, dtype=np.float32, copy=False)

        if value.size < 4 or value.size % 4 != 0:
            raise ValueError('Invalid arbp parameter: {}'.format(value))

        return value.ravel('C')


    def __getAttrTexUnit(self, attr):
        """Returns the texture unit identifier which corresponds to the named
        vertex attribute.
        """

        pos     = self.attrPositions[attr]
        texUnit = 'GL_TEXTURE{}'.format(pos)
        texUnit = getattr(gl, texUnit)

        return texUnit


    def __generatePositions(self, textureMap=None):
        """Called by :meth:`__init__`. Generates positions for vertex/fragment
        program parameters and vertex attributes.

        The lengths of each vertex/fragment parameter are known (see
        :mod:`.arbp.parse`), so these parameters are set up to be sequentially
        stored in the program parameter memory.

        Vertex attributes are passed to the vertex program as texture
        coordinates.

        If texture units were not specified in ``__init__``, texture units are
        also automatically assigned to each texture used in the fragment
        program.
        """

        vpPoses   = {}
        fpPoses   = {}
        texPoses  = {}
        attrPoses = {}

        # Vertex parameters
        pos = 0
        for name in self.vertParams:
            vpPoses[name]  = pos
            pos           += self.vertParamLens[name]

        # Fragment parameters
        pos = 0
        for name in self.fragParams:
            fpPoses[name]  = pos
            pos           += self.fragParamLens[name]

        # Vertex attributes
        for i, name in enumerate(self.attrs):
            attrPoses[name]  = i

        # Texture positions. If the caller did
        # not provide a texture map in __init__,
        # we'll generate some positions.
        if textureMap is None:

            names    = self.textures
            poses    = list(range(len(names)))
            texPoses = {n : p for n, p in zip(names, poses)}
        else:
            texPoses = dict(textureMap)

        return vpPoses, fpPoses, texPoses, attrPoses


    def __cleanSource(self, src):
        """Strips out comments and blank lines from the given string, unless
        ``clean is False`` (as passed into :meth:`__init__`).
        """

        if not self.clean:
            return src

        # strip out comments and blank lines
        lines = src.split('\n')
        lines = [l.strip() for l in lines]
        lines = [l         for l in lines if l    != '']
        lines = [l         for l in lines if l[0] != '#']
        src  = '\n'.join(lines)

        # Squeeze duplicate spaces
        src = re.sub(' +', ' ', src)

        return src


    def __compile(self, vertSrc, fragSrc):
        """Called by :meth:`__init__`. Compiles the vertex and fragment
        programs and returns references to the compiled programs.

        """

        gl.glEnable(arbvp.GL_VERTEX_PROGRAM_ARB)
        gl.glEnable(arbfp.GL_FRAGMENT_PROGRAM_ARB)

        # Clear out unnecessary stuff from
        # the source, and make sure it is
        # plain ASCII - not unicode.
        vertSrc = self.__cleanSource(vertSrc)
        fragSrc = self.__cleanSource(fragSrc)
        vertSrc = vertSrc.encode('ascii')
        fragSrc = fragSrc.encode('ascii')

        fragProg = arbfp.glGenProgramsARB(1)
        vertProg = arbvp.glGenProgramsARB(1)

        # vertex program
        try:
            arbvp.glBindProgramARB(arbvp.GL_VERTEX_PROGRAM_ARB, vertProg)
            arbvp.glProgramStringARB(arbvp.GL_VERTEX_PROGRAM_ARB,
                                     arbvp.GL_PROGRAM_FORMAT_ASCII_ARB,
                                     len(vertSrc),
                                     vertSrc)

        except Exception:

            position = gl.glGetIntegerv(arbvp.GL_PROGRAM_ERROR_POSITION_ARB)
            message  = gl.glGetString(  arbvp.GL_PROGRAM_ERROR_STRING_ARB)
            message  = message.decode('ascii')

            raise RuntimeError('Error compiling vertex program ({}): '
                               '{}\n{}'.format(position, message, vertSrc))

        # fragment program
        try:
            arbfp.glBindProgramARB(arbfp.GL_FRAGMENT_PROGRAM_ARB,
                                   fragProg)

            arbfp.glProgramStringARB(arbfp.GL_FRAGMENT_PROGRAM_ARB,
                                     arbfp.GL_PROGRAM_FORMAT_ASCII_ARB,
                                     len(fragSrc),
                                     fragSrc)
        except Exception:
            position = gl.glGetIntegerv(arbfp.GL_PROGRAM_ERROR_POSITION_ARB)
            message  = gl.glGetString(  arbfp.GL_PROGRAM_ERROR_STRING_ARB)
            message  = message.decode('ascii')

            raise RuntimeError('Error compiling fragment program ({}): '
                               '{}\n{}'.format(position, message, fragSrc))

        gl.glDisable(arbvp.GL_VERTEX_PROGRAM_ARB)
        gl.glDisable(arbfp.GL_FRAGMENT_PROGRAM_ARB)

        return vertProg, fragProg



#!/usr/bin/env python
#
# overlaylistpanel.py - The OverlayListPanel.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the ``OverlayListPanel``, a *FSLeyes control* which
displays a list of all overlays currently in the :class:`.OverlayList`.
"""


import logging
import colorsys

import          wx
import numpy as np

import fsl.data.image                as fslimage
import fsl.utils.idle                as idle

import fsleyes_props                 as props
import fsleyes_widgets.bitmaptoggle  as bmptoggle
import fsleyes_widgets.elistbox      as elistbox

import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.icons                 as icons
import fsleyes.autodisplay           as autodisplay
import fsleyes.strings               as strings
import fsleyes.tooltips              as fsltooltips
import fsleyes.actions.loadoverlay   as loadoverlay
import fsleyes.actions.saveoverlay   as saveoverlay
import fsleyes.actions.removeoverlay as removeoverlay


log = logging.getLogger(__name__)


class OverlayListPanel(ctrlpanel.ControlPanel):
    """The ``OverlayListPanel`` displays all overlays in the
    :class:`.OverlayList`, and allows the user to add, remove, and re-order
    overlays. An ``OverlayListPanel`` looks something like this:

    .. image:: images/overlaylistpanel.png
       :scale: 50%
       :align: center


    A :class:`ListItemWidget` is displayed alongside every overlay in the
    list - this allows the user to enable/disable, group, and save each
    overlay.

    The ``OverlayListPanel`` is closely coupled to a few
    :class:`.DisplayContext` properties: the
    :attr:`.DisplayContext.selectedOverlay` property is linked to the currently
    selected item in the overlay list, and the order in which the overlays are
    shown is defined by the :attr:`.DisplayContext.overlayOrder` property. This
    property is updated when the user changes the order of items in the list.
    """


    @staticmethod
    def defaultLayout():
        """Returns a dictionary containing layout settings to be passed to
        :class:`.ViewPanel.togglePanel`.
        """
        return {'location' : wx.BOTTOM}


    def __init__(self,
                 parent,
                 overlayList,
                 displayCtx,
                 viewPanel,
                 showVis=True,
                 showGroup=True,
                 showSave=True,
                 propagateSelect=True,
                 elistboxStyle=None,
                 filterFunc=None):
        """Create an ``OverlayListPanel``.

        :arg parent:          The :mod:`wx` parent object.

        :arg overlayList:     An :class:`.OverlayList` instance.

        :arg displayCtx:      A :class:`.DisplayContext` instance.

        :arg viewPanel:       The :class:`.ViewPanel` instance.

        :arg showVis:         If ``True`` (the default), a button will be shown
                              alongside each overlay, allowing the user to
                              toggle the overlay visibility.

        :arg showGroup:       If ``True`` (the default), a button will be shown
                              alongside each overlay, allowing the user to
                              toggle overlay grouping.

        :arg showSave:        If ``True`` (the default), a button will be shown
                              alongside each overlay, allowing the user to save
                              the overlay (if it is not saved).

        :arg propagateSelect: If ``True`` (the default), when the user
                              interacts with the :class:`.ListItemWidget` for
                              an overlay which is *not* the currently selected
                              overlay, that overlay is updated to be the
                              selected overlay.

        :arg elistboxStyle:   Style flags passed through to the
                              :class:`.EditableListBox`.

        :arg filterFunc:      Function which must accept an overlay as its
                              sole argument, and return ``True`` or ``False``.
                              If this function returns ``False`` for an
                              overlay, the :class:`ListItemWidget` for that
                              overlay will be disabled.
        """

        def defaultFilter(o):
            return True

        if filterFunc is None:
            filterFunc = defaultFilter

        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, viewPanel)

        self.__showVis         = showVis
        self.__showGroup       = showGroup
        self.__showSave        = showSave
        self.__propagateSelect = propagateSelect
        self.__filterFunc      = filterFunc

        if elistboxStyle is None:
            elistboxStyle = (elistbox.ELB_REVERSE      |
                             elistbox.ELB_TOOLTIP_DOWN |
                             elistbox.ELB_SCROLL_BUTTONS)

        # list box containing the list of overlays - it
        # is populated in the _overlayListChanged method
        self.__listBox = elistbox.EditableListBox(self, style=elistboxStyle)

        # listeners for when the user does
        # something with the list box
        self.__listBox.Bind(elistbox.EVT_ELB_SELECT_EVENT,   self.__lbSelect)
        self.__listBox.Bind(elistbox.EVT_ELB_MOVE_EVENT,     self.__lbMove)
        self.__listBox.Bind(elistbox.EVT_ELB_REMOVE_EVENT,   self.__lbRemove)
        self.__listBox.Bind(elistbox.EVT_ELB_ADD_EVENT,      self.__lbAdd)
        self.__listBox.Bind(elistbox.EVT_ELB_DBLCLICK_EVENT, self.__lbDblClick)

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(self.__sizer)

        self.__sizer.Add(self.__listBox, flag=wx.EXPAND, proportion=1)

        self.overlayList.addListener(
            'overlays',
            self.name,
            self.__overlayListChanged)

        self.displayCtx.addListener(
            'overlayOrder',
            self.name,
            self.__overlayListChanged)

        self.displayCtx.addListener(
            'selectedOverlay',
            self.name,
            self.__selectedOverlayChanged)

        self.__overlayListChanged()

        self.__selectedOverlayChanged()

        self.Layout()

        self.__minSize = self.__sizer.GetMinSize()
        self.SetMinSize(self.__minSize)


    def GetMinSize(self):
        """Returns the minimum size for this ``OverlayListPanel``.

        Under Linux/GTK, the ``wx.agw.lib.aui`` layout manager seems to
        arbitrarily adjust the minimum sizes of some panels. Therefore, The
        minimum size of the ``OverlayListPanel`` is calculated in
        :meth:`__init__`, and is fixed.
        """
        return self.__minSize


    def destroy(self):
        """Must be called when this ``OverlayListPanel`` is no longer needed.
        Removes some property listeners, and calls
        :meth:`.ControlPanel.destroy`.
        """

        self.overlayList.removeListener('overlays',        self.name)
        self.displayCtx .removeListener('selectedOverlay', self.name)
        self.displayCtx .removeListener('overlayOrder',    self.name)

        # A listener on name was added
        # in the _overlayListChanged method
        for overlay in self.overlayList:
            display = self.displayCtx.getDisplay(overlay)
            display.removeListener('name', self.name)

        self.__filterFunc = None
        self.__listBox.Clear()

        ctrlpanel.ControlPanel.destroy(self)


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` property
        changes. Updates the selected item in the list box.
        """

        if len(self.overlayList) > 0:
            self.__listBox.SetSelection(
                self.displayCtx.getOverlayOrder(
                    self.displayCtx.selectedOverlay))


    def __overlayNameChanged(self, value, valid, display, propName):
        """Called when the :attr:`.Display.name` of an overlay changes. Updates
        the corresponding label in the overlay list.
        """

        overlay = display.overlay
        idx     = self.displayCtx.getOverlayOrder(overlay)
        name    = display.name

        if name is None:
            name = ''

        self.__listBox.SetItemLabel(idx, name)


    def __overlayListChanged(self, *a):
        """Called when the :class:`.OverlayList` changes.  All of the items
        in the overlay list are re-created.
        """

        self.__listBox.Clear()

        for i, overlay in enumerate(self.displayCtx.getOrderedOverlays()):

            display  = self.displayCtx.getDisplay(overlay)
            name     = display.name
            if name is None: name = ''

            self.__listBox.Append(name, overlay)

            widget = ListItemWidget(self,
                                    overlay,
                                    display,
                                    self.displayCtx,
                                    self.__listBox,
                                    showVis=self.__showVis,
                                    showGroup=self.__showGroup,
                                    showSave=self.__showSave,
                                    propagateSelect=self.__propagateSelect)

            if not self.__filterFunc(overlay):
                widget.Disable()

            self.__listBox.SetItemWidget(i, widget)

            tooltip = overlay.dataSource
            if tooltip is None:
                tooltip = strings.labels['OverlayListPanel.noDataSource']
            self.__listBox.SetItemTooltip(i, tooltip)

            display.addListener('name',
                                self.name,
                                self.__overlayNameChanged,
                                overwrite=True)

        if len(self.overlayList) > 0:
            self.__listBox.SetSelection(
                self.displayCtx.getOverlayOrder(
                    self.displayCtx.selectedOverlay))

        self.__listBox.Layout()


    def __lbMove(self, ev):
        """Called when an overlay is moved in the :class:`.EditableListBox`.
        Reorders the :attr:`.DisplayContext.overlayOrder` to reflect the
        change.
        """
        self.displayCtx.disableListener('overlayOrder', self.name)
        self.displayCtx.overlayOrder.move(ev.oldIdx, ev.newIdx)
        self.displayCtx.enableListener('overlayOrder', self.name)


    def __lbSelect(self, ev):
        """Called when an overlay is selected in the
        :class:`.EditableListBox`. Updates the
        :attr:`.DisplayContext.selectedOverlay` property.
        """


        self.displayCtx.disableListener('selectedOverlay', self.name)
        self.displayCtx.selectedOverlay = self.displayCtx.overlayOrder[ev.idx]
        self.displayCtx.enableListener('selectedOverlay', self.name)


    def __lbAdd(self, ev):
        """Called when the *add* button on the list box is pressed.
        Calls the :func:`.loadoverlay.interactiveLoadOverlays` method.
        """

        def onLoad(paths, overlays):
            if len(overlays) == 0:
                return

            self.overlayList.extend(overlays)
            self.displayCtx.selectedOverlay = len(self.overlayList) - 1

            if self.displayCtx.autoDisplay:
                for overlay in overlays:
                    autodisplay.autoDisplay(overlay,
                                            self.overlayList,
                                            self.displayCtx)

        loadoverlay.interactiveLoadOverlays(
            onLoad=onLoad,
            inmem=self.displayCtx.loadInMemory)


    def __lbRemove(self, ev):
        """Called when an item is removed from the overlay listbox.
        Removes the corresponding overlay from the :class:`.OverlayList`.
        """

        overlay = self.displayCtx.overlayOrder[ev.idx]
        overlay = self.overlayList[overlay]

        with props.skip(self.overlayList, 'overlays',     self.name), \
             props.skip(self.displayCtx,  'overlayOrder', self.name):

            if not removeoverlay.removeOverlay(self.overlayList,
                                               self.displayCtx,
                                               overlay):
                ev.Veto()

            # The overlayListChanged method
            # must be called asynchronously,
            # otherwise it will corrupt the
            # EditableListBox state
            else:
                idle.idle(self.__overlayListChanged)


    def __lbDblClick(self, ev):
        """Called when an item label is double clicked on the overlay list
        box. Toggles the visibility of the overlay, via the
        :attr:`.Display.enabled` property..
        """
        idx             = self.displayCtx.overlayOrder[ev.idx]
        overlay         = self.overlayList[idx]
        display         = self.displayCtx.getDisplay(overlay)
        display.enabled = not display.enabled


class ListItemWidget(wx.Panel):
    """A ``LisItemWidget`` is created by the :class:`OverlayListPanel` for
    every overlay in the :class:`.OverlayList`. A ``LisItemWidget`` contains
    controls which allow the user to:

     - Toggle the visibility of the overlay (via the :attr:`.Display.enabled`
       property)

     - Add the overlay to a group (see the
       :attr:`.DisplayContext.overlayGroups` property, and the :mod:`.group`
       module).

     - Save the overlay (if it has been modified).

    .. note:: While the :class:`.DisplayContext` allows multiple
              :class:`.OverlayGroup` instances to be defined (and added to its
              :attr:`.DisplayContext.overlayGroups` property), *FSLeyes*
              currently only defines a single group . This ``OverlayGroup``
              is created in the :func:`.fsleyes.context` function, and overlays
              can be added/removed to/from it via the *lock* button on a
              ``ListItemWidget``. This functionality might change in a future
              version of *FSLeyes*.


    .. note:: Currently, only :class:`.Image` overlays can be saved. The *save*
              button is disabled for all other overlay types.
    """


    def __init__(self,
                 parent,
                 overlay,
                 display,
                 displayCtx,
                 listBox,
                 showVis=True,
                 showGroup=True,
                 showSave=True,
                 propagateSelect=True):
        """Create a ``ListItemWidget``.

        :arg parent:          The :mod:`wx` parent object.
        :arg overlay:         The overlay associated with this
                              ``ListItemWidget``.
        :arg display:         The :class:`.Display` associated with the
                              overlay.
        :arg displayCtx:      The :class:`.DisplayContext` instance.
        :arg listBox:         The :class:`.EditableListBox` that contains this
                              ``ListItemWidget``.
        :arg showVis:         If ``True`` (the default), a button will be shown
                              allowing the user to toggle the overlay
                              visibility.
        :arg showGroup:       If ``True`` (the default), a button will be shown
                              allowing the user to toggle overlay grouping.
        :arg showSave:        If ``True`` (the default), a button will be shown
                              allowing the user to save the overlay (if it is
                              not saved).
        :arg propagateSelect: If ``True`` (the default), when an overlay is
                              selected in the list, the
                              :attr:`.DisplayContext.selectedOverlay` is
                              updated accordingly.
        """
        wx.Panel.__init__(self, parent)

        self.enabledFG  = wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOWTEXT)
        self.disabledFG = wx.SystemSettings.GetColour(wx.SYS_COLOUR_GRAYTEXT)

        # give unsaved overlays  reddish colours
        if wx.SystemSettings.GetAppearance().IsDark():
            self.unsavedDefaultBG  = '#443333'
            self.unsavedSelectedBG = '#551111'
        else:
            self.unsavedDefaultBG  = '#ffeeee'
            self.unsavedSelectedBG = '#ffcdcd'

        self.__overlay         = overlay
        self.__display         = display
        self.__displayCtx      = displayCtx
        self.__listBox         = listBox
        self.__propagateSelect = propagateSelect
        self.__name            = '{}_{}'.format(self.__class__.__name__,
                                                id(self))
        self.__sizer           = wx.BoxSizer(wx.HORIZONTAL)

        self.SetSizer(self.__sizer)
        self.Bind(wx.EVT_WINDOW_DESTROY, self.__onDestroy)

        btnStyle = wx.BU_EXACTFIT | wx.BU_NOTEXT

        if showSave:
            self.__saveButton = wx.Button(self, style=btnStyle)
            self.__saveButton.SetBitmapLabel(icons.loadBitmap('floppydisk16'))
            self.__saveButton.SetToolTip(
                wx.ToolTip(fsltooltips.actions[self, 'save']))

            self.__sizer.Add(self.__saveButton, flag=wx.EXPAND, proportion=1)

            if isinstance(overlay, fslimage.Image):
                overlay.register(self.__name,
                                 self.__saveStateChanged,
                                 'saveState')
                self.__saveButton.Bind(wx.EVT_BUTTON, self.__onSaveButton)
            else:
                self.__saveButton.Enable(False)

            self.__saveStateChanged()

        if showGroup:
            self.__lockButton = bmptoggle.BitmapToggleButton(
                self, style=btnStyle)

            self.__lockButton.SetBitmap(
                icons.loadBitmap('chainlinkHighlight16'),
                icons.loadBitmap('chainlink16'))

            self.__lockButton.SetToolTip(
                wx.ToolTip(fsltooltips.actions[self, 'group']))

            self.__sizer.Add(self.__lockButton, flag=wx.EXPAND, proportion=1)

            # There is currently only one overlay
            # group in the application. In the
            # future there may be multiple groups.
            group = displayCtx.overlayGroups[0]

            group  .addListener('overlays',
                                self.__name,
                                self.__overlayGroupChanged)

            self.__lockButton.Bind(wx.EVT_TOGGLEBUTTON, self.__onLockButton)
            self.__overlayGroupChanged()

        # Set up the show/hide button if needed
        if showVis:
            self.__visibility = bmptoggle.BitmapToggleButton(
                self,
                trueBmp=icons.loadBitmap('eyeHighlight16'),
                falseBmp=icons.loadBitmap('eye16'),
                style=btnStyle)

            self.__visibility.SetToolTip(
                wx.ToolTip(fsltooltips.properties[display, 'enabled']))

            self.__sizer.Add(self.__visibility, flag=wx.EXPAND, proportion=1)

            display.addListener('enabled',
                                self.__name,
                                self.__displayVisChanged)

            self.__visibility.Bind(bmptoggle.EVT_BITMAP_TOGGLE,
                                   self.__onVisButton)

            self.__displayVisChanged()


    def __overlayGroupChanged(self, *a):
        """Called when the :class:`.OverlayGroup` changes. Updates the *lock*
        button based on whether the overlay associated with this
        ``ListItemWidget`` is in the group or not.
        """
        group = self.__displayCtx.overlayGroups[0]
        self.__lockButton.SetValue(self.__overlay in group.overlays)


    def __onSaveButton(self, ev):
        """Called when the *save* button is pushed. Calls the
        :meth:`.Image.save` method.
        """

        if self.__propagateSelect:
            self.__displayCtx.selectOverlay(self.__overlay)

        if not self.__overlay.saveState:
            saveoverlay.saveOverlay(self.__overlay, self.__display)


    def __onLockButton(self, ev):
        """Called when the *lock* button is pushed. Adds/removes the overlay
        to/from the :class:`.OverlayGroup`.
        """
        if self.__propagateSelect:
            self.__displayCtx.selectOverlay(self.__overlay)

        group = self.__displayCtx.overlayGroups[0]

        if self.__lockButton.GetValue(): group.addOverlay(   self.__overlay)
        else:                            group.removeOverlay(self.__overlay)


    def __onVisButton(self, ev):
        """Called when the *visibility* button is pushed. Toggles the overlay
        visibility.
        """

        if self.__propagateSelect:
            self.__displayCtx.selectOverlay(self.__overlay)

        idx     = self.__listBox.IndexOf(self.__overlay)
        enabled = self.__visibility.GetValue()

        with props.suppress(self.__display, 'enabled', self.__name):
            self.__display.enabled = enabled

        if enabled: fgColour = self.enabledFG
        else:       fgColour = self.disabledFG

        self.__listBox.SetItemForegroundColour(idx, fgColour)


    def __displayVisChanged(self, *a):
        """Called when the :attr:`.Display.enabled` property of the overlay
        changes. Updates the state of the *enabled* buton, and changes the
        item foreground colour.
        """

        idx = self.__listBox.IndexOf(self.__overlay)

        enabled = self.__display.enabled

        if enabled: fgColour = self.enabledFG
        else:       fgColour = self.disabledFG

        self.__visibility.SetValue(enabled)
        self.__listBox.SetItemForegroundColour(idx, fgColour)


    def __onDestroy(self, ev):
        """Called when this ``ListItemWidget`` is destroyed (i.e. when the
        associated overlay is removed from the :class:`OverlayListPanel`).
        Removes some proprety listeners from the :class:`.Display` and
        :class:`.OverlayGroup` instances, and from the overlay if it is an
        :class:`.Image` instance.
        """
        ev.Skip()
        if ev.GetEventObject() is not self:
            return

        group = self.__displayCtx.overlayGroups[0]

        if self.__display.hasListener('enabled',  self.__name):
            self.__display.removeListener('enabled',  self.__name)

        if group.hasListener('overlays', self.__name):
            group.removeListener('overlays', self.__name)

        if isinstance(self.__overlay, fslimage.Image):

            # Notifier.deregister will ignore
            # non-existent listener de-registration
            self.__overlay.deregister(self.__name, 'saveState')


    def __saveStateChanged(self, *a):
        """If the overlay is an :class:`.Image` instance, this method is
        called when its :attr:`.Image.saved` property changes. Updates the
        state of the *save* button.
        """

        if not isinstance(self.__overlay, fslimage.Image):
            return

        idx = self.__listBox.IndexOf(self.__overlay)

        self.__saveButton.Enable(not self.__overlay.saveState)

        if self.__overlay.saveState:
            self.__listBox.SetItemBackgroundColour(idx)

        else:
            self.__listBox.SetItemBackgroundColour(
                idx,
                self.unsavedDefaultBG,
                self.unsavedSelectedBG),

        tooltip = self.__overlay.dataSource
        if tooltip is None:
            tooltip = strings.labels['OverlayListPanel.noDataSource']
        self.__listBox.SetItemTooltip(idx, tooltip)



#!/usr/bin/env python
#
# reloadoverlay.py - The ReloadOverlayAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ReloadOverlayAction`, a global action
which reloads the currently selected overlay from disk.
"""


import logging
import os.path as op



import fsl.data.image               as fslimage
import fsleyes_widgets.utils.status as status
import fsleyes_props                as props
from . import                          base
from . import                          removeoverlay


log = logging.getLogger(__name__)


class ReloadOverlayAction(base.Action):
    """The ``ReloadOverlayAction`` reloads the currently selected overlay
    from disk. Currently only :class:`.Image` overlays are supported.
    """

    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``ReloadOverlayAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.Action.__init__(
            self, overlayList, displayCtx, self.__reloadOverlay)

        self.__frame = frame
        self.__name  = '{}_{}'.format(type(self).__name__, id(self))

        overlayList.addListener('overlays',
                                self.__name,
                                self.__selectedOverlayChanged)
        displayCtx .addListener('selectedOverlay',
                                self.__name,
                                self.__selectedOverlayChanged)

        self.__selectedOverlayChanged()


    def destroy(self):
        """Must be called when this ``ReloadOverlayAction`` is no longer
        required. Removes some property listeners, and calls
        :meth:`.Action.destroy`.
        """
        self.overlayList.removeListener('overlays',        self.__name)
        self.displayCtx .removeListener('selectedOverlay', self.__name)
        base.Action.destroy(self)


    def __selectedOverlayChanged(self, *a):
        """Called when the currently selected overlay changes. Enables/disables
        this ``Action`` depending on the type of the newly selected overlay.
        """
        ovl          = self.displayCtx.getSelectedOverlay()
        self.enabled = (ovl            is not None)    and \
                       (ovl.dataSource is not None)    and \
                       type(ovl) == fslimage.Image     and \
                       op.exists(ovl.dataSource)


    def __reloadOverlay(self):
        """Reloads the currently selected overlay from disk.
        """
        ovl = self.displayCtx.getSelectedOverlay()

        if ovl is None or type(ovl) != fslimage.Image:
            raise RuntimeError('Only Image overlays can be reloaded')

        index      = self.overlayList.index(ovl)
        order      = self.displayCtx.overlayOrder[:]
        dataSource = ovl.dataSource

        status.update('Reloading {}...'.format(dataSource))

        # Get refs to all DisplayContexts -
        # the master one, and the one for
        # every view panel.
        displayCtxs  = [self.displayCtx]
        viewPanels   = self.__frame.viewPanels
        displayCtxs += [vp.displayCtx for vp in viewPanels]

        # Now get refs to all Display and
        # DisplayOpts instances for this
        # overlay.
        displays = []
        opts     = []

        for dctx in displayCtxs:
            displays.append(self.displayCtx.getDisplay(ovl))
            opts    .append(self.displayCtx.getOpts(   ovl))

        # Turn those references into
        # {prop : value} dictionaries
        for i in range(len(displays)):

            d = displays[i]
            o = opts[    i]

            displayProps = d.getAllProperties()[0]
            optProps     = o.getAllProperties()[0]

            displays[i] = {p : getattr(d, p) for p in displayProps}
            opts[    i] = {p : getattr(o, p) for p in optProps}

        # Now that we've got all the settings
        # for this overlay, we'll remove it
        # from the list. If removeOverlay
        # returns False, it probably means
        # the user cancelled the action.
        if not removeoverlay.removeOverlay(self.overlayList,
                                           self.displayCtx,
                                           ovl,
                                           'reloadoverlay.unsaved'):
            return

        # Now we re-load the overlay, and add it
        # back in to the list at the same location
        ovl = fslimage.Image(dataSource)
        self.overlayList.insert(index, ovl)

        # Make sure the overlay is selected,
        # and the display order is preserved
        self.displayCtx.selectOverlay(ovl)
        self.displayCtx.overlayOrder = order

        # The last step is to re-apply all of the
        # Display/DisplayOpts settings to the
        # newly created Display/DisplayOpts
        # instances.
        for i, dctx in enumerate(displayCtxs):

            displayProps = displays[i]
            optProps     = opts[    i]
            d            = dctx.getDisplay(ovl)

            for prop, val in displayProps.items():

                if not d.propertyIsEnabled(prop):
                    continue

                try:
                    setattr(d, prop, val)
                except props.DisabledError:
                    continue

            # Get a ref to the DisplayOpts instance
            # after we have configured the Display,
            # as its overlay type may have changed
            # (and hence the DisplayOpts instance
            # may have been re-created).
            o = dctx.getOpts(ovl)
            for prop, val in optProps.items():

                if not o.propertyIsEnabled(prop):
                    continue

                setattr(o, prop, val)

        status.update('{} reloaded.'.format(dataSource))



#!/usr/bin/env python
#
# cropimage.py - The CropImagePanel class
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`CropImagePanel` class.

The ``CropImagePanel`` is a a FSLeyes control which is used in conjunction
with the :class:`.OrthoCropProfile`, allowing the user to crop an image.
This module also provides the standalone :func:`loadCropParameters` function,
for loading cropping parameters from a file.
"""


import              os
import os.path   as op
import itertools as it
import              wx
import numpy     as np

import fsl.utils.idle                            as idle
import fsl.data.image                            as fslimage

import fsleyes_props                             as props
import fsleyes_widgets.rangeslider               as rslider
import fsleyes_widgets.utils.status              as status

import fsleyes.controls.controlpanel             as ctrlpanel
import fsleyes.displaycontext                    as displaycontext
import fsleyes.views.orthopanel                  as orthopanel
import fsleyes.strings                           as strings
import fsleyes.actions                           as actions
import fsleyes.actions.copyoverlay               as copyoverlay
import fsleyes.controls.displayspacewarning      as dswarning
import fsleyes.plugins.profiles.orthocropprofile as orthocropprofile


class CropImageAction(actions.ToggleControlPanelAction):
    """The ``CropImageAction`` just toggles a :class:`.CropImagePanel`. It is
    added under the FSLeyes Tools menu.
    """

    @staticmethod
    def supportedViews():
        """The ``CropImageAction`` is restricted for use with
        :class:`.OrthoPanel` views.
        """
        return [orthopanel.OrthoPanel]


    def __init__(self, overlayList, displayCtx, ortho):
        """Create  ``CropImageAction``. """
        super().__init__(overlayList, displayCtx, ortho, CropImagePanel)
        self.__ortho = ortho
        self.__name  = '{}_{}'.format(type(self).__name__, id(self))

        displayCtx.addListener('selectedOverlay', self.__name,
                               self.__selectedOverlayChanged)


    def destroy(self):
        """Called when the :class:`.OrthoPanel` that owns this action is
        closed. Clears references, removes listeners, and calls the base
        class ``destroy`` method.
        """
        if self.destroyed:
            return

        self.__ortho = None
        self.displayCtx.removeListener('selectedOverlay', self.__name)
        super().destroy()


    def __selectedOverlayChanged(self, *a):
        """Called when the selected overlay changes. Enables/disables this
        action (and hence the bound Tools menu item) depending on whether the
        overlay is an image.
        """
        ovl = self.displayCtx.getSelectedOverlay()
        self.enabled = isinstance(ovl, fslimage.Image)


class CropImagePanel(ctrlpanel.ControlPanel):
    """The ``CropImagePanel`` class is a FSLeyes control for use in an
    :class:`.OrthoPanel`, with the associated :class:`.CropImageProfile`.  It
    contains controls allowing the user to define a cropping box for the
    currently selected overlay (if it is an :class:`.Image`), and "Crop",
    "Load", "Save", and "Cancel" buttons.
    """


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``CropImagePanel`` is only intended to be added to
        :class:`.OrthoPanel` views.
        """
        from fsleyes.views.orthopanel import OrthoPanel
        return [OrthoPanel]


    @staticmethod
    def ignoreControl():
        """Tells FSLeyes not to add the ``CropImagePanel`` as an option to
        the Settings menu. Instead, the :class:`CropImageAction` is added as
        an option to the Tools menu.
        """
        return True


    @staticmethod
    def profileCls():
        """Returns the :class:`.OrthoCropProfile` class, which needs to be
        activated in conjunction with the ``CropImagePanel``.
        """
        return orthocropprofile.OrthoCropProfile


    @staticmethod
    def defaultLayout():
        """Returns a dictionary containing layout settings to be passed to
        :class:`.ViewPanel.togglePanel`.
        """
        return {'floatPane' : True,
                'floatOnly' : True}



    def __init__(self, parent, overlayList, displayCtx, ortho):
        """Create a ``CropImagePanel``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg ortho:       The :class:`.OrthoPanel` instance.
        """
        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, ortho)

        profile = ortho.currentProfile

        self.__ortho   = ortho
        self.__profile = profile
        self.__overlay = None

        self.__cropBoxWidget = props.makeWidget(
            self,
            profile,
            'cropBox',
            showLimits=False,
            labels=['xmin', 'xmax', 'ymin', 'ymax', 'zmin', 'zmax'])

        self.__volumeWidget = rslider.RangeSliderSpinPanel(
            self,
            minValue=0,
            maxValue=1,
            minDistance=1,
            lowLabel='tmin',
            highLabel='tmax',
            style=rslider.RSSP_INTEGER)

        self.__dsWarning = dswarning.DisplaySpaceWarning(
            self,
            overlayList,
            displayCtx,
            ortho.frame,
            strings.messages[self, 'dsWarning'],
            'not like overlay',
            'overlay')

        self.__cropLabel       = wx.StaticText(self)
        self.__sizeLabel       = wx.StaticText(self)
        self.__cropButton      = wx.Button(    self, id=wx.ID_OK)
        self.__robustFovButton = wx.Button(    self)
        self.__loadButton      = wx.Button(    self)
        self.__saveButton      = wx.Button(    self)
        self.__cancelButton    = wx.Button(    self, id=wx.ID_CANCEL)

        self.__cropButton     .SetLabel(strings.labels[self, 'crop'])
        self.__robustFovButton.SetLabel(strings.labels[self, 'robustFov'])
        self.__loadButton     .SetLabel(strings.labels[self, 'load'])
        self.__saveButton     .SetLabel(strings.labels[self, 'save'])
        self.__cancelButton   .SetLabel(strings.labels[self, 'cancel'])

        self.__sizer    = wx.BoxSizer(wx.VERTICAL)
        self.__btnSizer = wx.BoxSizer(wx.HORIZONTAL)

        self.__sizer.Add((1, 10))
        self.__sizer.Add(self.__cropLabel,     flag=wx.CENTRE)
        self.__sizer.Add((1, 10))
        self.__sizer.Add(self.__dsWarning,     flag=wx.CENTRE)
        self.__sizer.Add((1, 10), proportion=1)
        self.__sizer.Add(self.__cropBoxWidget, flag=wx.EXPAND)
        self.__sizer.Add(self.__volumeWidget,  flag=wx.EXPAND)
        self.__sizer.Add((1, 10))
        self.__sizer.Add(self.__sizeLabel,     flag=wx.CENTRE, proportion=1)
        self.__sizer.Add((1, 10))
        self.__sizer.Add(self.__btnSizer,      flag=wx.CENTRE)
        self.__sizer.Add((1, 10))

        self.__btnSizer.Add((10, 1),                flag=wx.EXPAND,
                            proportion=1)
        self.__btnSizer.Add(self.__cropButton,      flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),                flag=wx.EXPAND)
        self.__btnSizer.Add(self.__robustFovButton, flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),                flag=wx.EXPAND)
        self.__btnSizer.Add(self.__loadButton,      flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),                flag=wx.EXPAND)
        self.__btnSizer.Add(self.__saveButton,      flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),                flag=wx.EXPAND)
        self.__btnSizer.Add(self.__cancelButton,    flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),                flag=wx.EXPAND,
                            proportion=1)

        self.SetSizer(self.__sizer)
        self.SetMinSize(self.__sizer.GetMinSize())
        self.__cropButton.SetDefault()

        self.__cropButton  .Bind(wx.EVT_BUTTON,          self.__onCrop)
        self.__loadButton  .Bind(wx.EVT_BUTTON,          self.__onLoad)
        self.__saveButton  .Bind(wx.EVT_BUTTON,          self.__onSave)
        self.__cancelButton.Bind(wx.EVT_BUTTON,          self.__onCancel)
        self.__volumeWidget.Bind(rslider.EVT_RANGE,      self.__onVolume)
        self.__volumeWidget.Bind(rslider.EVT_LOW_RANGE,  self.__onVolume)
        self.__volumeWidget.Bind(rslider.EVT_HIGH_RANGE, self.__onVolume)

        profile.robustfov.bindToWidget(self,
                                       wx.EVT_BUTTON,
                                       self.__robustFovButton)

        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__selectedOverlayChanged)
        overlayList.addListener('overlays',
                                self.name,
                                self.__selectedOverlayChanged)
        profile    .addListener('cropBox',
                                self.name,
                                self.__cropBoxChanged)

        self.__selectedOverlayChanged()
        self.__cropBoxChanged()


    def destroy(self):
        """Must be called when this ``CropImagePanel`` is no longer needed.
        Removes property listeners and clears references.
        """

        profile     = self.__profile
        displayCtx  = self.displayCtx
        overlayList = self.overlayList
        dsWarning   = self.__dsWarning

        profile    .removeListener('cropBox',         self.name)
        displayCtx .removeListener('selectedOverlay', self.name)
        overlayList.removeListener('overlays',        self.name)

        self.__ortho     = None
        self.__profile   = None
        self.__dsWarning = None

        dsWarning.destroy()
        ctrlpanel.ControlPanel.destroy(self)


    def __registerOverlay(self, overlay):
        """Called by :meth:`__selectedOverlayChanged`. Registers the
        given overlay.
        """

        self.__overlay = overlay

        display = self.displayCtx.getDisplay(overlay)
        is4D    = overlay.ndim >= 4

        if is4D:
            self.__volumeWidget.SetLimits(0, overlay.shape[3])
            self.__volumeWidget.SetRange( 0, overlay.shape[3])

        self.__volumeWidget.Enable(is4D)

        display.addListener('name', self.name, self.__overlayNameChanged)
        self.__overlayNameChanged()


    def __deregisterOverlay(self):
        """Called by :meth:`__selectedOverlayChanged`. Deregisters the
        current overlay.
        """

        if self.__overlay is None:
            return

        try:
            display = self.displayCtx.getDisplay(self.__overlay)
            display.removeListener('name', self.name)

        except displaycontext.InvalidOverlayError:
            pass

        self.__cropLabel.SetLabel(strings.labels[self, 'image.noImage'])

        self.__overlay = None


    def __overlayNameChanged(self, *a):
        """Called when the :attr:`.Display.name` of the currently selected
        overlay changes. Updates the name label.
        """

        display = self.displayCtx.getDisplay(self.__overlay)
        label   = strings.labels[self, 'image']
        label   = label.format(display.name)
        self.__cropLabel.SetLabel(label)


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` changes.
        Updates labels appropriately.
        """

        displayCtx = self.displayCtx
        overlay    = displayCtx.getSelectedOverlay()

        if overlay is self.__overlay:
            return

        self.__deregisterOverlay()

        if not isinstance(overlay, fslimage.Image):
            self.Disable()
        else:
            self.Enable()
            self.__registerOverlay(overlay)


    def __updateSizeLabel(self):
        """Called by the crop region and volume widget event handlers. Updates
        a label which displays the current crop region size.
        """

        overlay = self.__overlay
        profile = self.__profile
        xlen    = profile.cropBox.xlen
        ylen    = profile.cropBox.ylen
        zlen    = profile.cropBox.zlen
        tlo     = self.__volumeWidget.GetLow()
        thi     = self.__volumeWidget.GetHigh()
        tlen    = thi - tlo

        if overlay.ndim >= 4:
            label = strings.labels[self, 'cropSize4d']
            label = label.format(xlen, ylen, zlen, tlen)
        else:
            label = strings.labels[self, 'cropSize3d']
            label = label.format(xlen, ylen, zlen)

        self.__sizeLabel.SetLabel(label)


    def __cropBoxChanged(self, *a):
        """Called when the :attr:`.OrthoCropProfile.cropBox` changes.
        Updates labels appropriately.
        """
        self.__updateSizeLabel()


    def __onVolume(self, ev):
        """Called when the user changes the volume limit, for 4D images.
        Updates the label which displays the crop region size.
        """
        self.__updateSizeLabel()


    def __onLoad(self, ev):
        """Called when the Save button is pushed. Prompts the user to select
        a file to load crop parameters from.
        """

        overlay  = self.__overlay
        cropBox  = self.__profile.cropBox
        fileName = '{}_crop.txt'.format(overlay.name)

        if overlay.dataSource is not None:
            dirName = op.dirname(overlay.dataSource)
        else:
            dirName = os.getcwd()

        if not op.exists(op.join(dirName, fileName)):
            fileName = ''

        dlg = wx.FileDialog(
            self,
            defaultDir=dirName,
            defaultFile=fileName,
            message=strings.messages[self, 'saveCrop'],
            style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)

        if dlg.ShowModal() != wx.ID_OK:
            return

        filePath = dlg.GetPath()
        errTitle = strings.titles[  self, 'loadError']
        errMsg   = strings.messages[self, 'loadError']

        with status.reportIfError(errTitle, errMsg, raiseError=False):

            params     = loadCropParameters(filePath, overlay)
            cropBox[:] = params[:6]

            if overlay.ndim >= 4:
                tlo, thi = params[6:]
                self.__volumeWidget.SetLow(tlo)
                self.__volumeWidget.SetHigh(thi)


    def __onSave(self, ev):
        """Called when the Save button is pushed. Saves the current crop
        parameters to a text file.
        """

        overlay  = self.__overlay
        cropBox  = self.__profile.cropBox
        fileName = '{}_crop.txt'.format(overlay.name)

        if overlay.dataSource is not None:
            dirName = op.dirname(overlay.dataSource)
        else:
            dirName = os.getcwd()

        dlg = wx.FileDialog(
            self,
            defaultDir=dirName,
            defaultFile=fileName,
            message=strings.messages[self, 'saveCrop'],
            style=wx.FD_SAVE)

        if dlg.ShowModal() != wx.ID_OK:
            return

        filePath = dlg.GetPath()

        # The crop parameters are saved
        # in a fslroi-compatible manner.
        params = [cropBox.xlo,
                  cropBox.xhi - cropBox.xlo,
                  cropBox.ylo,
                  cropBox.yhi - cropBox.ylo,
                  cropBox.zlo,
                  cropBox.zhi - cropBox.zlo]

        if overlay.ndim >= 4:
            tlo     = self.__volumeWidget.GetLow()
            thi     = self.__volumeWidget.GetHigh()
            params.extend((tlo, thi - tlo))

        errTitle = strings.titles[  self, 'saveError']
        errMsg   = strings.messages[self, 'saveError']

        with status.reportIfError(errTitle, errMsg, raiseError=False):
            np.savetxt(filePath, [params], fmt='%i')


    def __onCancel(self, ev=None):
        """Called when the Cancel button is pushed. Calls
        :meth:`.OrthoPanel.togglePanel` - this will result in
        this ``CropImagePanel`` being destroyed.

        This method is also called programmatically from the :meth:`__onCrop`
        method after the image is cropped.
        """

        # Do asynchronously, because we don't want
        # this CropImagePanel being destroyed from
        # its own event handler.
        idle.idle(self.__ortho.togglePanel, CropImagePanel)


    def __onCrop(self, ev):
        """Crops the selected image. This is done via a call to
        :func:`.copyoverlay.copyImage`. Also calls :meth:`__onCancel`,
        to finish cropping.
        """

        overlayList = self.overlayList
        displayCtx  = self.displayCtx
        overlay     = displayCtx.getSelectedOverlay()
        display     = displayCtx.getDisplay(overlay)
        name        = '{}_roi'.format(display.name)
        cropBox     = self.__profile.cropBox
        roi         = [cropBox.x,
                       cropBox.y,
                       cropBox.z]

        if overlay.ndim >= 4:
            roi.append(self.__volumeWidget.GetRange())

        copyoverlay.copyImage(
            overlayList,
            displayCtx,
            overlay,
            createMask=False,
            copy4D=True,
            copyDisplay=True,
            name=name,
            roi=roi)

        self.__onCancel()


def loadCropParameters(filename, overlay):
    """Load in crop values from a text file assumed to contain ``fslroi``-
    compatible parameters. Any parameters which may be passed to ``fslroi``
    are accepted::

        fslroi in out tmin tlen
        fslroi in out xmin xlen ymin ylen zmin zlen
        fslroi in out xmin xlen ymin ylen zmin zlen tmin tlen

    Any of the ``len`` parameters may be equal to -1, in which case it is
    interpreted as continuing from the low index

    :arg filename: File to load crop parameters from.
    :arg overlay:  An :class:`.Image` which is the cropping target.
    :returns:      A sequence of ``lo, hi`` crop parameters.
    """

    is4D   = overlay.ndim >= 4
    shape  = overlay.shape[:4]
    params = list(np.loadtxt(filename).flatten())

    if len(params) not in (2, 6, 8):
        raise ValueError('File contains the wrong number of crop parameters')
    if len(params) in (2, 8) and not is4D:
        raise ValueError('File contains the wrong number of crop parameters')

    if len(params) == 2:
        params = [0, -1, 0, -1, 0, -1] + params

    if is4D and len(params) == 6:
        params = params + [0, -1]

    los = []
    his = []

    for dim in range(len(shape)):

        dlo  = params[dim * 2]
        dlen = params[dim * 2 + 1]

        if dlen == -1:
            dlen = shape[dim] - dlo

        dhi = dlo + dlen

        los.append(dlo)
        his.append(dhi)

    for lo, hi, lim in zip(los, his, shape):
        if lo < 0 or hi > lim:
            raise ValueError('Crop parameters are out of bounds for image '
                             'shape ({} < 0 or {} > {}'.format(lo, hi, lim))

    return list(it.chain(*zip(los, his)))



#!/usr/bin/env python
#
# lightboxpanel.py - A panel which contains a LightBoxCanvas, for displaying
# multiple slices from a collection of overlays.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LightBoxPanel`, which displays multiple
2D slices of 3D overlays.
"""

import logging

import wx

import numpy as np

import fsleyes_widgets.utils.layout         as fsllayout

import fsleyes.actions                      as actions
import fsleyes.gl.wxgllightboxcanvas        as lightboxcanvas
import fsleyes.profiles.lightboxviewprofile as lightboxviewprofile
import fsleyes.displaycontext.lightboxopts  as lightboxopts
from . import                                  canvaspanel


log = logging.getLogger(__name__)


class LightBoxPanel(canvaspanel.CanvasPanel):
    """The ``LightBoxPanel`` is a *FSLeyes view* which is capable of
    displaying multiple 2D slices of the 3D overlays conatined in an
    :class:`.OverlayList`. A ``LightBoxPanel`` looks something like the
    following:

    .. image:: images/lightboxpanel.png
       :scale: 50%
       :align: center


    The ``LightBoxPanel`` uses a :class:`.LightBoxCanvas` panel to display
    the slices, and a :class:`.LightBoxOpts` instance to manage the display
    settings. The canvas is accessed through the :meth:`getCanvas` and
    :meth:`getGLCanvases` methods, and the ``LightBoxOpts`` instanace can
    be retrieved via the :meth:`.CanvasPanel.sceneOpts` property.


    The ``LightBoxPanel`` adds the following actions to those already
    provided by the :class:`.CanvasPanel`:

    .. autosummary::
       :nosignatures:

       toggleLightBoxToolBar
    """


    @staticmethod
    def defaultLayout():
        """Returns a list of control panel types to be added for the default
        lightbox panel layout.
        """
        return ['OverlayDisplayToolBar',
                'LightBoxToolBar',
                'OverlayListPanel',
                'LocationPanel']


    @staticmethod
    def controlOrder():
        """Returns a list of control panel names, specifying the order in
        which they should appear in the  FSLeyes ortho panel settings menu.
        """
        return ['OverlayListPanel',
                'LocationPanel',
                'OverlayInfoPanel',
                'OverlayDisplayPanel',
                'CanvasSettingsPanel',
                'AtlasPanel',
                'OverlayDisplayToolBar',
                'LigbhtBoxToolBar',
                'FileTreePanel']


    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create a ``LightBoxPanel``.

        :arg parent:      A :mod:`wx` parent object.
        :arg overlayList: A :class:`.OverlayList` instance.
        :arg displayCtx:  A :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        """

        sceneOpts = lightboxopts.LightBoxOpts(self)

        canvaspanel.CanvasPanel.__init__(self,
                                         parent,
                                         overlayList,
                                         displayCtx,
                                         frame,
                                         sceneOpts)

        self.__scrollbar = wx.ScrollBar(self.centrePanel, style=wx.SB_VERTICAL)
        self.__lbCanvas  = lightboxcanvas.WXGLLightBoxCanvas(
            self.contentPanel,
            overlayList,
            displayCtx)

        lbopts = self.__lbCanvas.opts

        lbopts.bindProps('pos', displayCtx, 'location')
        lbopts.bindProps('zax',             sceneOpts)
        lbopts.bindProps('bgColour',        sceneOpts)
        lbopts.bindProps('cursorColour',    sceneOpts)
        lbopts.bindProps('showCursor',      sceneOpts)
        lbopts.bindProps('showGridLines',   sceneOpts)
        lbopts.bindProps('highlightSlice',  sceneOpts)
        lbopts.bindProps('renderMode',      sceneOpts)
        lbopts.bindProps('highDpi',         sceneOpts)

        # Bind these properties the other way around,
        # so that the sensible values calcualted by
        # the LBCanvas during its initialisation are
        # propagated to the LBOpts instance, rather
        # than the non-sensible default values in the
        # LBOpts instance.
        sceneOpts.bindProps('nrows',        lbopts)
        sceneOpts.bindProps('ncols',        lbopts)
        sceneOpts.bindProps('topRow',       lbopts)
        sceneOpts.bindProps('sliceSpacing', lbopts)
        sceneOpts.bindProps('zrange',       lbopts)

        self.__canvasSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.contentPanel.SetSizer(self.__canvasSizer)

        self.__canvasSizer.Add(self.__lbCanvas, flag=wx.EXPAND, proportion=1)

        self.displayCtx .addListener('selectedOverlay',
                                     self.name,
                                     self.__selectedOverlayChanged)
        self.displayCtx .addListener('displaySpace',
                                     self.name,
                                     self.__radioOrientationChanged)
        self.displayCtx .addListener('radioOrientation',
                                     self.name,
                                     self.__radioOrientationChanged)
        self.overlayList.addListener('overlays',
                                     self.name,
                                     self.__selectedOverlayChanged)

        # When any lightbox properties change,
        # make sure the scrollbar is updated
        sceneOpts.addListener(
            'ncols',        self.name, self.__ncolsChanged)
        sceneOpts.addListener(
            'nrows',        self.name, self.__onLightBoxChange)
        sceneOpts.addListener(
            'topRow',       self.name, self.__onLightBoxChange)
        sceneOpts.addListener(
            'sliceSpacing', self.name, self.__onLightBoxChange)
        sceneOpts.addListener(
            'zrange',       self.name, self.__onLightBoxChange)
        sceneOpts.addListener(
            'zax',          self.name, self.__onLightBoxChange)
        sceneOpts.addListener(
            'zoom',         self.name, self.__onZoom)

        # When the scrollbar is moved,
        # update the canvas display
        self.__scrollbar.Bind(wx.EVT_SCROLL, self.__onScroll)

        self.Bind(wx.EVT_SIZE, self.__onResize)

        sceneOpts.zoom = 750

        self.__onLightBoxChange()
        self.__onZoom()

        self.__selectedOverlayChanged()
        self.centrePanelLayout()
        self.initProfile(lightboxviewprofile.LightBoxViewProfile)


    def destroy(self):
        """Must be called when this ``LightBoxPanel`` is closed.

        Removes property listeners, destroys the :class:`.LightBoxCanvas`,
        and calls :meth:`.CanvasPanel.destroy`.
        """

        self.displayCtx .removeListener('selectedOverlay',  self.name)
        self.displayCtx .removeListener('displaySpace',     self.name)
        self.displayCtx .removeListener('radioOrientation', self.name)
        self.overlayList.removeListener('overlays',         self.name)

        canvaspanel.CanvasPanel.destroy(self)

        self.__lbCanvas.destroy()
        self.__lbCanvas = None


    def getActions(self):
        """Overrides :meth:`.ActionProvider.getActions`. Returns all of the
        :mod:`.actions` that are defined on this ``LightBoxPanel``.
        """
        actionz = [self.screenshot,
                   self.showCommandLineArgs,
                   self.applyCommandLineArgs,
                   None,
                   self.toggleMovieMode,
                   self.toggleDisplaySync]

        names = [a.actionName if a is not None else None for a in actionz]
        return list(zip(names, actionz))


    def getGLCanvases(self):
        """Returns a list containing the :class:`.LightBoxCanvas` contained
        within this ``LightBoxPanel``.
        """
        return [self.__lbCanvas]


    def getCanvas(self):
        """Returns a reference to the :class:`.LightBoxCanvas` instance. """
        return self.__lbCanvas


    def centrePanelLayout(self):
        """Overrides :meth:`.CanvasPanel.centrePanelLayout`. Adds the
        scrollbar to the centre panel.
        """

        self.layoutContainerPanel()

        centrePanel    = self.centrePanel
        containerPanel = self.containerPanel
        sizer          = wx.BoxSizer(wx.HORIZONTAL)

        centrePanel.SetSizer(sizer)

        sizer.Add(containerPanel,   flag=wx.EXPAND, proportion=1)
        sizer.Add(self.__scrollbar, flag=wx.EXPAND)

        self.PostSizeEvent()


    def __radioOrientationChanged(self, *a):
        """Called when the :attr:`.DisplayContext.displaySpace` or
        :attr:`.DisplayContext.radioOrientation` properties change. Updates the
        :attr:`.LightBoxCanvas.invertX` property as needed.
        """

        lbopts  = self.__lbCanvas.opts
        inRadio = self.displayCtx.displaySpaceIsRadiological()
        flip    = self.displayCtx.radioOrientation != inRadio
        flip    = flip and lbopts.zax in (1, 2)

        lbopts.invertX = flip


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` changes.

        If the currently selected overlay is a :class:`.Nifti` instance, or
        has an associated reference image (see
        :meth:`.DisplayOpts.referenceImage`), a listener is registered on
        the reference image :attr:`.NiftiOpts.transform` property, so that the
        :meth:`__transformChanged` method will be called when it changes.
        """

        if len(self.overlayList) == 0:
            return

        selectedOverlay = self.displayCtx.getSelectedOverlay()

        for overlay in self.overlayList:

            refImage = self.displayCtx.getReferenceImage(overlay)

            if refImage is None:
                continue

            opts = self.displayCtx.getOpts(refImage)

            opts.removeListener('transform', self.name)

            if overlay == selectedOverlay:
                opts.addListener('transform',
                                 self.name,
                                 self.__transformChanged)

        # If the current zrange is [0, 0]
        # we'll assume that the spacing/
        # zrange need to be initialised.
        lbCanvas = self.__lbCanvas
        opts     = self.sceneOpts

        if opts.zrange == [0.0, 0.0]:

            opts.sliceSpacing = lbCanvas.calcSliceSpacing(selectedOverlay)
            opts.zrange       = self.displayCtx.bounds.getRange(opts.zax)


    def __transformChanged(self, *a):
        """Called when the :attr:`.NiftiOpts.transform` property for the
        reference image of the currently selected overlay changes.

        Updates the :attr:`.LightBoxOpts.sliceSpacing` and
        :attr:`.LightBoxOpts.zrange` properties to values sensible to the
        new overlay display space.
        """

        sceneOpts = self.sceneOpts
        overlay   = self.displayCtx.getReferenceImage(
            self.displayCtx.getSelectedOverlay())

        if overlay is None:
            return

        opts     = self.displayCtx.getOpts(overlay)
        loBounds = opts.bounds.getLo()
        hiBounds = opts.bounds.getHi()

        # Reset the spacing/zrange. Not
        # sure if this is the best idea,
        # but it's here for the time being.
        sceneOpts.sliceSpacing = self.__lbCanvas.calcSliceSpacing(overlay)
        sceneOpts.zrange.x     = (loBounds[sceneOpts.zax],
                                  hiBounds[sceneOpts.zax])

        self.__onResize()


    def __onZoom(self, *a):
        """Called when the :attr:`.SceneOpts.zoom` property changes. Updates
        the number of slice columns shown.
        """
        opts       = self.sceneOpts
        minval     = opts.getAttribute('zoom', 'minval')
        maxval     = opts.getAttribute('zoom', 'maxval')
        normZoom   = 1.0 - (opts.zoom - minval) / float(maxval)
        opts.ncols = int(1 + np.round(normZoom * 29))


    def __onResize(self, ev=None):
        """Called when the panel is resized. Automatically adjusts the number
        of rows to the maximum displayable number (given that the number of
        columns is fixed).
        """

        if ev is not None:
            ev.Skip()

        if self.destroyed:
            return

        # Lay this panel out, so the
        # canvas panel size is up to date
        self.Layout()

        lbcanvas          = self.__lbCanvas
        lbopts            = lbcanvas.opts
        width,   height   = lbcanvas        .GetClientSize().Get()
        sbWidth, sbHeight = self.__scrollbar.GetClientSize().Get()

        width = width - sbWidth

        xlen = self.displayCtx.bounds.getLen(lbopts.xax)
        ylen = self.displayCtx.bounds.getLen(lbopts.yax)

        sliceWidth  = width / float(lbopts.ncols)
        sliceHeight = fsllayout.calcPixHeight(xlen, ylen, sliceWidth)

        if sliceHeight > 0:
            lbopts.nrows = int(height / sliceHeight)


    def __ncolsChanged(self, *a):
        """Called when the :attr:`.LightBoxOpts.ncols` property changes.
        Calculates the number of rows to display, and updates the
        scrollbar.
        """
        self.__onResize()
        self.__onLightBoxChange()


    def __onLightBoxChange(self, *a):
        """Called when any :class:`.LightBoxOpts` property changes.

        Updates the scrollbar to reflect the change.
        """
        canvas = self.__lbCanvas
        opts   = canvas.opts
        self.__scrollbar.SetScrollbar(opts.topRow,
                                      opts.nrows,
                                      canvas.getTotalRows(),
                                      opts.nrows,
                                      True)


    def __onScroll(self, *a):
        """Called when the scrollbar is moved.

        Updates the top row displayed on the :class:`.LightBoxCanvas`.
        """
        self.__lbCanvas.opts.topRow = self.__scrollbar.GetThumbPosition()



#!/usr/bin/env python
#
# powerspectrumpanel.py - The PowerSpectrumPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`PowerSpectrumPanel` class, a
:class:`.ViewPanel` which plots frequency/power spectra.
"""


import logging

import numpy as np

import fsl.data.image                       as fslimage
import fsl.data.mesh                        as fslmesh
import fsl.data.melodicimage                as fslmelimage
import fsleyes_props                        as props

import fsleyes.actions                      as actions
import fsleyes.views.plotpanel              as plotpanel
import fsleyes.profiles.plotprofile         as plotprofile
import fsleyes.plotting.powerspectrumseries as psseries


log = logging.getLogger(__name__)


class PowerSpectrumPanel(plotpanel.OverlayPlotPanel):
    """The ``PowerSpectrumPanel`` class is an :class:`.OverlayPlotPanel` which
    plots power spectra of overlay data.  ``PowerSpectrumPanel`` uses
    :class:`.PowerSpectrumSeries` to plot the power spectra of overlay data.


    A couple of control panels may be shown on a ``PowerSpectrumPanel`` via
    :meth:`.ViewPanel.togglePanel`.

    .. autosummary::
       :nosignatures:

       ~fsleyes.controls.plotlistpanel.PlotListPanel
       ~fsleyes.controls.powerspectrumcontrolpanel.PowerSpectrumControlPanel


    **Melodic images**


    The :class:`.PowerSpectrumSeries` class uses a fourier transform to
    calculate the power spectrum of a time course.  However,
    :class:`.MelodicImage` overlays already have an associated power spectrum,
    meaning that there is no need to calculate one for them..  So for these
    overlays, a :class:`.MelodicPowerSpectrumSeries` instance is used.
    """


    plotMelodicICs  = props.Boolean(default=True)
    """If ``True``, the power spectra of :class:`.MelodicImage` overlays are
    plotted using :class:`.MelodicPowerSpectrumSeries` instances. Otherwise,
    :class:`.MelodicImage` overlays are treated as regular :class:`.Image`
    overlays, and :class:`.VoxelPowerSpectrumSeries` are used for plotting.
    """


    plotFrequencies = props.Boolean(default=True)
    """If ``True``, the x axis is scaled so that it represents frequency. """


    @staticmethod
    def defaultLayout():
        """Returns a list of control panel types to be added for the default
        power spectrum panel layout.
        """
        return ['PowerSpectrumToolBar',
                'OverlayListPanel',
                'PlotListPanel']



    @staticmethod
    def controlOrder():
        """Returns a list of control panel names, specifying the order in
        which they should appear in the  FSLeyes ortho panel settings menu.
        """
        return ['OverlayListPanel',
                'PlotListPanel',
                'PowerSpectrumToolBar',
                'PowerSpectrumControlPanel']


    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create a ``PowerSpectrumPanel``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """

        plotpanel.OverlayPlotPanel.__init__(self,
                                            parent,
                                            overlayList,
                                            displayCtx,
                                            frame)

        self.addListener('plotFrequencies', self.name, self.draw)
        self.addListener('plotMelodicICs',
                                self.name,
                                self.__plotMelodicICsChanged)

        self.initProfile(plotprofile.PlotProfile)


    def destroy(self):
        """Must be called when this ``PowerSpectrumPanel`` is no longer
        needed. Removes some property listeners, and calls
        :meth:`.OverlayPlotPanel.destroy`.
        """

        self.removeListener('plotFrequencies', self.name)
        self.removeListener('plotMelodicICs',  self.name)
        plotpanel.OverlayPlotPanel.destroy(self)


    def getActions(self):
        """Overrides :meth:`.ActionProvider.getActions`. Returns all of the
        :mod:`.actions` that are defined on this ``PowerSpectrumPanel``.
        """
        actionz = [self.screenshot,
                   self.importDataSeries,
                   self.exportDataSeries]

        names = [a.actionName if a is not None else None for a in actionz]
        return list(zip(names, actionz))


    def draw(self, *a):
        """Overrides :meth:`.PlotPanel.draw`. Draws some
        :class:`.PowerSpectrumSeries` using the
        :meth:`.PlotPanel.drawDataSeries` method.
        """

        if not self or self.destroyed:
            return

        canvas = self.canvas
        pss    = self.getDataSeriesToPlot()

        for ps in pss:
            with props.suppress(ps, 'label'):
                ps.label = ps.makeLabel()

        canvas.drawDataSeries(extraSeries=pss)
        canvas.drawArtists()


    def createDataSeries(self, overlay):
        """Overrides :meth:`.OverlayPlotPanel.createDataSeries`. Creates a
        :class:`.PowerSpectrumSeries` instance for the given overlay.
        """

        displayCtx  = self.displayCtx
        overlayList = self.overlayList

        psargs = [overlay, overlayList, displayCtx, self.canvas]

        if isinstance(overlay, fslmesh.Mesh):
            ps        = psseries.MeshPowerSpectrumSeries(*psargs)
            opts      = displayCtx.getOpts(overlay)
            targets   = [displayCtx, opts]
            propNames = ['location', 'vertexData']

        elif isinstance(overlay, fslmelimage.MelodicImage) and \
             self.plotMelodicICs:

            ps        = psseries.MelodicPowerSpectrumSeries(*psargs)
            targets   = [displayCtx.getOpts(overlay)]
            propNames = ['volume']

        elif isinstance(overlay, fslimage.Image) and overlay.ndim > 3:

            if overlay.iscomplex:
                ps = psseries.ComplexPowerSpectrumSeries(*psargs)
            else:
                ps = psseries.VoxelPowerSpectrumSeries(*psargs)

            opts      = displayCtx.getOpts(overlay)
            targets   = [displayCtx, opts]
            propNames = ['location', 'volumeDim']

        else:
            return None, None, None

        ps.colour    = self.getOverlayPlotColour(overlay)
        ps.lineStyle = self.getOverlayPlotStyle(overlay)
        ps.lineWidth = 2
        ps.alpha     = 1.0

        return ps, targets, propNames


    def prepareDataSeries(self, ps):
        """Overrides :meth:`.PlotPanel.prepareDataSeries`.  Performs some
        pre-processing on the data of the given :class:`.PowerSpectrumSeries`
        instance.
        """

        xdata, ydata = ps.getData()

        # if plotFrequencies is disabled, replace
        # the x axis data with sample numbers
        if (xdata is not None) and \
           (ydata is not None) and \
           (len(xdata) > 0)    and \
           not self.plotFrequencies:
            xdata = np.arange(len(ydata))

        return xdata, ydata


    def __plotMelodicICsChanged(self, *a):
        """Called when the :attr:`plotMelodicICs` property changes. Re-creates
        the internally cached :class:`.PowerSpectrmSeries` instances for all
        :class:`.MelodicImage` overlays in the :class:`.OverlayList`.
        """

        for overlay in self.overlayList:
            if isinstance(overlay, fslmelimage.MelodicImage):
                self.clearDataSeries(overlay)

        self.updateDataSeries()
        self.draw()



#!/usr/bin/env python
#
# loadstandard.py - Action which allows the user to open standard images.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LoadStandardAction`, which allows the user
to load in standard space images from the ``$FSLDIR/data/standard/`` directory.
"""


import os.path as op

import fsl.utils.platform as fslplatform
from . import                loadoverlay
from . import                base


class LoadStandardAction(base.Action):
    """The ``LoadStandardAction`` prompts the user to open one or more
    overlays, using ``$FSLDIR/data/standard/`` as the default directory.
    This functionality is provided in the :mod:`.loadoverlay` module.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``LoadStandardAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.Action.__init__(
            self, overlayList, displayCtx, self.__loadStandard)

        self.__name = '{}_{}'.format(type(self).__name__, id(self))
        self.__setStandardDir()

        # If FSLDIR is not set, the setStandardDir
        # disables this action. But we'll listen
        # for changes to FSLDIR, in case it gets
        # set later on.
        fslplatform.platform.register(self.__name, self.__setStandardDir)


    def destroy(self):
        """Must be called when this ``LoadStandardAction`` is no longer
        needed. Performs some clean-up.
        """
        fslplatform.platform.deregister(self.__name)
        base.Action.destroy(self)


    def __setStandardDir(self, *a):
        """Called by :meth:`__init__`, and when the
        :attr:`~fsl.utils.Platform.fsldir` property is changed. Updates
        the path to the FSLDIR standard directory.
        """

        fsldir = fslplatform.platform.fsldir

        if fsldir is not None:
            self.__stddir = op.join(fsldir, 'data', 'standard')
        else:
            self.__stddir = None

        self.enabled = self.__stddir is not None


    def __loadStandard(self):
        """Calls the :func:`.loadoverlay.interactiveLoadOverlays` method.
        If the user added some overlays, updates the
        :attr:`.DisplayContext.selectedOverlay` accordingly.
        """

        def onLoad(paths, overlays):

            if len(overlays) == 0:
                return

            self.overlayList.extend(overlays)
            self.displayCtx.selectedOverlay = self.displayCtx.overlayOrder[-1]

        loadoverlay.interactiveLoadOverlays(
            fromDir=self.__stddir,
            onLoad=onLoad,
            inmem=self.displayCtx.loadInMemory)



#!/usr/bin/env python
#
# projectimagetosurface.py - the ProjectImageToSurfaceAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ProjectImageToSurfaceAction` class,
which allows data from an :class:`.Image` overlay to be projected onto
a :class:`.Mesh` overlay.
"""

import                  wx
import numpy         as np
import scipy.ndimage as ndi

import fsl.transform.affine as affine
import fsl.data.mesh        as fslmesh
import fsl.data.image       as fslimage
import fsleyes.strings      as strings
import fsleyes.actions.base as base


class ProjectImageToSurfaceAction(base.NeedOverlayAction):
    """The ``ProjectImageToSurfaceAction`` class allows the user to project
    data from a volumetric :class:`.Image` overlay onto a :class:`.Mesh`
    overlay.

    A ``ProjectImageToSurfaceAction`` is active when the currently selected
    overlay is a :class:`.Mesh`. When executed, the user is prompted to
    select an :class:`.Image` overlay to project onto the mesh. Only images
    which overlap the bounding box of the ``Mesh`` are available as options.

    When the user selects an :class:`.Image`, the data from the image at each
    vertex in the mesh is retrieved using ``scipy.ndimage.map_coordinates``.
    This data is then added as an option on the :attr:`.MeshOpts.vertexData`
    property, and selected.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``ProjectImageToSurfaceAction``.

        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext`
        :arg frame:       The :class:`.ViewPanel` this action is associated
                          with.
        """
        super().__init__(overlayList,
                         displayCtx,
                         self.__projectImage,
                         fslmesh.Mesh)
        self.__frame = frame


    def __projectImage(self):
        """Run the ``ProjectImageToSurfaceAction``. """

        # TODO Allow user to load an image from file?

        # We need to use child DisplayOpts objects in
        # order to perform the projection, as the parent
        # DisplayOpts objects do not maintain information
        # about transformations between the different
        # coordinate systems. Any child DisplayContext
        # will do.
        displayCtx = self.__frame.focusedViewPanel.displayCtx
        mesh       = self.displayCtx.getSelectedOverlay()
        mopts      = displayCtx.getOpts(mesh)

        # any images which overlap with the
        # mesh bbox in the display coordinate
        # system are given as options
        images = []
        mbbox  = affine.transform(
            mesh.bounds, mopts.getTransform('mesh', 'display'))
        mbbox  = list(zip(*mbbox))

        for o in self.overlayList:

            if not isinstance(o, fslimage.Image):
                continue

            iopts = displayCtx.getOpts(o)
            ibbox = affine.axisBounds(
                o.shape[:3], iopts.getTransform('voxel', 'display'))
            ibbox = list(zip(*ibbox))

            if overlap(mbbox, ibbox):
                images.append(o)

        # can't find any images which
        # overlap with the mesh in the
        # display coordinate system
        if len(images) == 0:
            wx.MessageDialog(
                self.__frame,
                message=strings.messages[self, 'noOverlap'],
                style=(wx.ICON_EXCLAMATION | wx.OK)).ShowModal()
            return

        # ask the user what image
        # they want to project
        dlg = wx.SingleChoiceDialog(
            self.__frame,
            caption=strings.titles[self, 'dialog'],
            message=strings.labels[self, 'message'],
            style=wx.OK | wx.CANCEL,
            choices=[i.name for i in images])

        if dlg.ShowModal() != wx.ID_OK:
            return

        # Sample data from the selected image
        image = images[dlg.GetSelection()]
        vdata = projectImageDataOntoMesh(displayCtx, image, mesh)

        # add the vertex data to
        # the mesh, and select it
        key              = mopts.addVertexData(image.name, vdata)
        mopts.vertexData = key


def projectImageDataOntoMesh(displayCtx, image, mesh):
    """Samples data from ``image`` at every vertex on ``mesh``. Uses
    ``scipy.ndimage.map_coordinates``.
    """

    # transform the mesh vertices into
    # the image voxel coordinate system
    mopts   = displayCtx.getOpts(mesh)
    iopts   = displayCtx.getOpts(image)
    m2d     = mopts.getTransform('mesh', 'display')
    verts   = mesh.vertices
    verts   = iopts.transformCoords(verts, 'display', 'voxel', pre=m2d)

    # sample the image data at
    # every vertex location
    imgdata = image.data[iopts.index()]
    vdata   = ndi.map_coordinates(imgdata,
                                  verts.T,
                                  output=np.float64,
                                  cval=np.nan)

    return vdata


def overlap(bbox1, bbox2):
    """Returns ``True`` if the two bounding boxes overlap at all, ``False``
    otherwise.

    Both ``bbox1`` and ``bbox2`` are expected to be sequences of
    ``(low, high)`` tuples containing the bounds of each axis in the
    coordinate system that the bounding boxes are defined in.
    """

    def axisOverlap(lo1, hi1, lo2, hi2):
        return (hi1 >= lo2) and (lo1 <= hi2) or \
               (hi2 >= lo1) and (lo2 <= hi1)

    for (lo1, hi1), (lo2, hi2) in zip(bbox1, bbox2):
        if not axisOverlap(lo1, hi1, lo2, hi2):
            return False

    return True



#!/usr/bin/env python
#
# timeseriescontrolpanel.py - The TimeSeriesControlPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`TimeSeriesControlPanel` a *FSLeyes
control* which allows the user to configure a :class:`.TimeSeriesPanel`.
"""


import wx

import fsleyes_props                 as props
import fsl.data.image                as fslimage
import fsl.data.featimage            as fslfeatimage

import fsleyes.tooltips              as fsltooltips
import fsleyes.plotting.timeseries   as timeseries
import fsleyes.strings               as strings
import fsleyes.views.timeseriespanel as timeseriespanel
from . import plotcontrolpanel       as plotctrl


class TimeSeriesControlPanel(plotctrl.PlotControlPanel):
    """The ``TimeSeriesControlPanel`` is a :class:`.PlotContrlPanel` which
    allows the user to configure a :class:`.TimeSeriesPanel`. It contains
    controls which are linked to the properties of the ``TimeSeriesPanel``,
    (which include properties defined on the :class:`.PlotPanel` base class),
    and the :class:`.DataSeries` class.


    A ``TimeSeriesControlPanel`` looks something like this:

    .. image:: images/timeseriescontrolpanel.png
       :scale: 50%
       :align: center


    The settings shown on a ``TimeSeriesControlPanel`` are organised into three
    or four sections:

     - The *Time series plot settings* section has controls which are linked to
       properties of the :class:`.PlotCanvas` class.

     - The *General plot settings* section has controls which are linked to
       properties of the :class:`.PlotPanel` base class.

     - The *Settings for the current time course* section has controls which
       are linked to properties of the :class:`.DataSeries` class. These
       properties define how the *current* time course is displayed (see the
       :class:`.TimeSeriesPanel` class documentation).

     - The *FEAT plot settings* is only shown if the currently selected overlay
       is a :class:`.FEATImage`. It has controls which are linked to properties
       of the :class:`.FEATTimeSeries` class.
    """


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``TimeSeriesControlPanel`` is only intended to be added to
        :class:`.TimeSeriesPanel` views.
        """

        return [timeseriespanel.TimeSeriesPanel]


    @staticmethod
    def defaultLayout():
        """Returns a dictionary containing layout settings to be passed to
        :class:`.ViewPanel.togglePanel`.
        """
        return {'location' : wx.RIGHT}


    def __init__(self, *args, **kwargs):
        """Create a ``TimeSeriesControlPanel``. All arguments are passed
        through to the :meth:`.PlotControlPanel.__init__` method.
        """

        plotctrl.PlotControlPanel.__init__(self, *args, **kwargs)

        self.plotPanel.addListener('plotMelodicICs',
                                   self.name,
                                   self.__plotMelodicICsChanged)


    def destroy(self):
        """Must be called when this ``TimeSeriesControlPanel`` is no longer
        needed. Removes some property listeners, and calls
        :meth:`.PlotControlPanel.destroy`.
        """
        psPanel = self.getPlotPanel()
        psPanel.removeListener('plotMelodicICs', self.name)
        plotctrl.PlotControlPanel.destroy(self)


    def generateCustomPlotPanelWidgets(self, groupName):
        """Overrides :meth:`.PlotControlPanel.generateCustomPlotPanelWidgets`.

        Adds some widgets for controlling some properties of the
        :class:`.TimeSeriesPanel`.
        """

        widgetList = self.getWidgetList()
        allWidgets = []
        tsPanel    = self.plotPanel
        tsProps    = ['plotMode',
                      'usePixdim',
                      'plotMelodicICs']

        for prop in tsProps:
            kwargs = {}
            if prop == 'plotMode':
                kwargs['labels'] = strings.choices[tsPanel, 'plotMode']

            widget = props.makeWidget(widgetList, tsPanel, prop, **kwargs)
            allWidgets.append(widget)
            widgetList.AddWidget(
                widget,
                displayName=strings.properties[tsPanel, prop],
                tooltip=fsltooltips.properties[tsPanel, prop],
                groupName=groupName)

        return allWidgets


    def generateCustomDataSeriesWidgets(self, ts, groupName):
        """Overrides :meth:`.PlotControlPanel.generateCustomDataSeriesWidgets`.

        Adds some widgets to the widget list for certain time series types.
        """

        overlay = ts.overlay

        if (isinstance(overlay, fslfeatimage.FEATImage)    and
            isinstance(ts,      timeseries.FEATTimeSeries) and
            overlay.hasStats()):
            return self.__generateFeatWidgets(ts, groupName)

        elif isinstance(ts, timeseries.ComplexTimeSeries):
            return self.__generateComplexWidgets(ts, groupName)

        else:
            return []


    def __generateFeatWidgets(self, ts, groupName):
        """Called by :meth:`generateCustomDataSeriesWidgets`. Generates
        widgets for :class:`.FEATTimeSeries` options, and adds them
        to the widget list.
        """

        overlay    = ts.overlay
        widgetList = self.getWidgetList()
        allWidgets = []

        full    = props.makeWidget(     widgetList, ts, 'plotFullModelFit')
        res     = props.makeWidget(     widgetList, ts, 'plotResiduals')
        evs     = props.makeListWidgets(widgetList, ts, 'plotEVs')
        pes     = props.makeListWidgets(widgetList, ts, 'plotPEFits')
        copes   = props.makeListWidgets(widgetList, ts, 'plotCOPEFits')
        partial = props.makeWidget(     widgetList, ts, 'plotPartial')
        data    = props.makeWidget(     widgetList, ts, 'plotData')

        allWidgets.append(full)
        allWidgets.append(res)
        allWidgets.extend(evs)
        allWidgets.extend(pes)
        allWidgets.extend(copes)
        allWidgets.append(partial)
        allWidgets.append(data)

        widgetList.AddWidget(
            data,
            displayName=strings.properties[ts, 'plotData'],
            tooltip=fsltooltips.properties[ts, 'plotData'],
            groupName=groupName)
        widgetList.AddWidget(
            full,
            displayName=strings.properties[ts, 'plotFullModelFit'],
            tooltip=fsltooltips.properties[ts, 'plotFullModelFit'],
            groupName=groupName)

        widgetList.AddWidget(
            res,
            displayName=strings.properties[ts, 'plotResiduals'],
            tooltip=fsltooltips.properties[ts, 'plotResiduals'],
            groupName=groupName)

        widgetList.AddWidget(
            partial,
            displayName=strings.properties[ts, 'plotPartial'],
            tooltip=fsltooltips.properties[ts, 'plotPartial'],
            groupName=groupName)

        widgetList.AddSpace(groupName=groupName)

        for i, ev in enumerate(evs):

            evName = overlay.evNames()[i]
            widgetList.AddWidget(
                ev,
                displayName=strings.properties[ts, 'plotEVs'].format(
                    i + 1, evName),
                tooltip=fsltooltips.properties[ts, 'plotEVs'],
                groupName=groupName)

        widgetList.AddSpace(groupName=groupName)

        for i, pe in enumerate(pes):
            evName = overlay.evNames()[i]
            widgetList.AddWidget(
                pe,
                displayName=strings.properties[ts, 'plotPEFits'].format(
                    i + 1, evName),
                tooltip=fsltooltips.properties[ts, 'plotPEFits'],
                groupName=groupName)

        widgetList.AddSpace(groupName=groupName)

        copeNames = overlay.contrastNames()
        for i, (cope, name) in enumerate(zip(copes, copeNames)):
            widgetList.AddWidget(
                cope,
                displayName=strings.properties[ts, 'plotCOPEFits'].format(
                    i + 1, name),
                tooltip=fsltooltips.properties[ts, 'plotCOPEFits'],
                groupName=groupName)

        return allWidgets


    def __generateComplexWidgets(self, ts, groupName):
        """Called by :meth:`generateCustomDataSeriesWidgets`. Generates
        widgets for :class:`.ComplexTimeSeries` options, and adds them
        to the widget list.
        """
        widgetList = self.getWidgetList()
        allWidgets = []

        if isinstance(ts, timeseries.ComplexTimeSeries):
            for propName in ['plotReal', 'plotImaginary',
                             'plotMagnitude', 'plotPhase']:
                widg = props.makeWidget(widgetList, ts, propName)
                widgetList.AddWidget(
                    widg,
                    displayName=strings.properties[ts, propName],
                    tooltip=fsltooltips.properties[ts, propName],
                    groupName=groupName)
                allWidgets.append(widg)

        return allWidgets


    def __plotMelodicICsChanged(self, *a):
        """Called when the :attr:`.TimeSeriesPanel.plotMelodicICs` property
        changes. If the current overlay is a :class:`.MelodicImage`,
        re-generates the widgets in the *current time course* section, as
        the :class:`.DataSeries` instance associated with the overlay may
        have been re-created.
        """
        self.refreshDataSeriesWidgets()
